#!/usr/bin/env python

# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Extract the optimal distributions for the following features of the population of neurons:
   soma: radius
   basal dendrites: n_neurites
   apical dendrites: n_neurites
   axons: n_neurites
"""

import argparse
from collections import OrderedDict
from collections import defaultdict
from collections import namedtuple
from itertools import chain
import os
import json
from json import encoder

from neurom import get, load_neurons, NeuriteType, stats
from neurom.io.utils import get_morph_files

encoder.FLOAT_REPR = lambda o: format(o, '.2f')

Limits = namedtuple('Limits', 'min, max')
Feature = namedtuple('Feature', 'name, limits')
Component = namedtuple('Component', 'name, features')


# Map feature names to data getter namer
FEATURE_MAP = {
    'radius': 'soma_radii',
    'number': 'number_of_neurites',
    'segment_length': 'segment_lengths',
    'initial_radius': 'trunk_origin_radii',
    # 'taper_rate': lambda n, kwargs: np.array(list(n.iter_segments(mm.segment_taper_rate,
    #                                                              **kwargs))),
}

# Map component name to filtering parameters where applicable
PARAM_MAP = {
    'basal_dendrite': {'neurite_type': NeuriteType.basal_dendrite},
    'apical_dendrite': {'neurite_type': NeuriteType.apical_dendrite},
    'axon': {'neurite_type': NeuriteType.axon},
    'soma': None
}


def extract_data(neurons, feature, params=None):
    """Extracts feature from a list of neurons
       and transforms the fitted distribution in the correct format.
       Returns the optimal distribution and corresponding parameters.
       Normal distribution params (mean, std)
       Exponential distribution params (loc, scale)
       Uniform distribution params (min, range)
    """
    if params is None:
        params = {}

    feature_data = [get(FEATURE_MAP[feature], n, **params) for n in neurons]
    feature_data = list(chain(*feature_data))
    return stats.optimal_distribution(feature_data)


def transform_header(mtype_name):
    """Add header to json output to wrap around distribution data.
    """
    head_dict = OrderedDict()

    head_dict["m-type"] = mtype_name
    head_dict["components"] = defaultdict(OrderedDict)

    return head_dict


def transform_package(mtype, files, components):
    """Put together header and list of data into one json output.
       feature_list contains all the information about the data to be extracted:
       features, feature_names, feature_components, feature_min, feature_max
    """
    data_dict = transform_header(mtype)
    neurons = load_neurons(files)

    for comp in components:
        params = PARAM_MAP[comp.name]
        for feature in comp.features:
            result = stats.fit_results_to_dict(
                extract_data(neurons, feature.name, params),
                feature.limits.min, feature.limits.max
            )

            # When the distribution is normal with sigma = 0
            # it will be replaced with constant
            if result['type'] == 'normal' and result['sigma'] == 0.0:
                replace_result = OrderedDict((('type', 'constant'),
                                              ('val', result['mu'])))
                result = replace_result

            data_dict["components"][comp.name].update({feature.name: result})

    return data_dict


def get_mtype_from_filename(filename, sep='_'):
    """Get mtype of a morphology file from file name

    Assumes file name has structure 'a/b/c/d/mtype_xyx.abc'
    """
    return os.path.basename(filename).split(sep)[0]


def get_mtype_from_directory(filename):
    """Get mtype of a morphology file from file's parent directory name

    Assumes file name has structure 'a/b/c/mtype/xyx.abc'
    """
    return os.path.split(os.path.dirname(filename))[-1]


MTYPE_GETTERS = {
    'filename': get_mtype_from_filename,
    'directory': get_mtype_from_directory
}


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description='Morphology fit distribution extractor',
        epilog='Note: Outputs json of the optimal distribution \
                and corresponding parameters.')

    parser.add_argument('datapaths',
                        nargs='+',
                        help='Morphology data directory paths')

    parser.add_argument('--mtype', choices=MTYPE_GETTERS.keys(),
                        help='Get mtype from filename or parent directory')

    return parser.parse_args()


# Note: new features require a data getter function in
# FEATURE_MAP
# TODO: Set this config from a JSON or YAML file
COMPONENTS = [
    Component('soma',
              [Feature('radius', Limits(None, None))]),
    Component('basal_dendrite',
              [
                  Feature('number', Limits(0, None)),
                  Feature('segment_length', Limits(0, None)),
                  Feature('initial_radius', Limits(0, None)),
                  # Feature('taper_rate', Limits(0, None))
              ]),
    Component('apical_dendrite',
              [
                  Feature('number', Limits(0, None)),
                  Feature('segment_length', Limits(0, None)),
                  Feature('initial_radius', Limits(0, None)),
                  # Feature('taper_rate', Limits(0, None))
              ]),
    Component('axon',
              [
                  Feature('number', Limits(0, None)),
                  Feature('segment_length', Limits(0, None)),
                  Feature('initial_radius', Limits(0, None)),
                  # Feature('taper_rate', Limits(0, None))
              ]),
]


if __name__ == '__main__':
    args = parse_args()

    data_dirs = args.datapaths

    mtype_getter = MTYPE_GETTERS.get(args.mtype, lambda f: 'UNKNOWN')

    for d in data_dirs:
        mtype_files = defaultdict(list)
        for f in get_morph_files(d):
            mtype_files[mtype_getter(f)].append(f)

        _results = [transform_package(mtype_, files_, COMPONENTS)
                    for mtype_, files_ in mtype_files.items()]

        for res in _results:
            print(json.dumps(res, indent=2), '\n')



#!/usr/bin/env python
# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Example for comparison of the same feature of multiple cells."""
import argparse
from pathlib import Path

import pylab as pl
import neurom as nm
from neurom.io.utils import get_morph_files


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description='Feature Comparison Between Different Cells')

    parser.add_argument('-d',
                        '--datapath',
                        help='Data directory')

    parser.add_argument('-o',
                        '--odir',
                        default='.',
                        help='Output path')

    parser.add_argument('-f',
                        '--features',
                        nargs='+',
                        help='List features separated by spaces')

    return parser.parse_args()


def stylize(ax, name, feature):
    """Stylization modifications to the plots
    """
    ax.set_ylabel(feature)
    ax.set_title(name, fontsize='small')


def histogram(neuron, feature, ax, bins=15, normed=True, cumulative=False):
    """
    Plot a histogram of the selected feature for the population of neurons.
    Plots x-axis versus y-axis on a scatter|histogram|binned values plot.

    Parameters :

        neurons : neuron list

        feature : str
        The feature of interest.

        bins : int
        Number of bins for the histogram.

        cumulative : bool
        Sets cumulative histogram on.

        ax : axes object
            the axes in which the plot is taking place
    """

    feature_values = nm.get(feature, neuron)
    # generate histogram
    ax.hist(feature_values, bins=bins, cumulative=cumulative, normed=normed)


def plot_feature(feature, cell):
    """Plot a feature
    """
    fig = pl.figure()
    ax = fig.add_subplot(111)

    if cell is not None:
        try:
            histogram(cell, feature, ax)
        except ValueError:
            pass
        stylize(ax, cell.name, feature)
    return fig


if __name__ == '__main__':
    args = parse_args()

    for morph_file in get_morph_files(args.datapath):
        nrn = nm.load_neuron(morph_file)

        for _feature in args.features:
            f = plot_feature(_feature, nrn)
            figname = "{0}_{1}.eps".format(_feature, nrn.name)
            f.savefig(Path(args.odir, figname))
            pl.close(f)



#!/usr/bin/env python
# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Easy analysis examples

These examples highlight most of the pre-packaged neurom.nm.get
morphometrics functionality.

"""

from __future__ import print_function
from pprint import pprint
import numpy as np
import neurom as nm


def stats(data):
    """Dictionary with summary stats for data

    Returns:
        dicitonary with length, mean, sum, standard deviation,\
            min and max of data
    """
    return {'len': len(data),
            'mean': np.mean(data),
            'sum': np.sum(data),
            'std': np.std(data),
            'min': np.min(data),
            'max': np.max(data)}


def pprint_stats(data):
    """Pretty print summary stats for data."""
    pprint(stats(data))


if __name__ == '__main__':

    filename = 'test_data/swc/Neuron.swc'

    #  load a neuron from an SWC file
    nrn = nm.load_neuron(filename)

    # Get some soma information
    # Soma radius and surface area
    print("Soma radius", nm.get('soma_radii', nrn)[0])
    print("Soma surface area", nm.get('soma_surface_areas', nrn)[0])

    # Get information about neurites
    # Most neurite data can be queried for a particular type of neurite.
    # The allowed types are members of the NeuriteType enumeration.
    # NEURITE_TYPES is a list of valid neurite types.

    # We start by calling methods for different neurite types separately
    # to warm up...

    # number of neurites
    print('Number of neurites (all):', nm.get('number_of_neurites', nrn)[0])
    print('Number of neurites (axons):',
          nm.get('number_of_neurites', nrn, neurite_type=nm.NeuriteType.axon)[0])
    print('Number of neurites (apical dendrites):',
          nm.get('number_of_neurites', nrn, neurite_type=nm.NeuriteType.apical_dendrite)[0])
    print('Number of neurites (basal dendrites):',
          nm.get('number_of_neurites', nrn, neurite_type=nm.NeuriteType.basal_dendrite)[0])

    # number of sections
    print('Number of sections:',
          nm.get('number_of_sections', nrn)[0])
    print('Number of sections (axons):',
          nm.get('number_of_sections', nrn, neurite_type=nm.NeuriteType.axon)[0])
    print('Number of sections (apical dendrites):',
          nm.get('number_of_sections', nrn, neurite_type=nm.NeuriteType.apical_dendrite)[0])
    print('Number of sections (basal dendrites):',
          nm.get('number_of_sections', nrn, neurite_type=nm.NeuriteType.basal_dendrite)[0])

    # number of sections per neurite
    print('Number of sections per neurite:',
          nm.get('number_of_sections_per_neurite', nrn))
    print('Number of sections per neurite (axons):',
          nm.get('number_of_sections_per_neurite', nrn, neurite_type=nm.NeuriteType.axon))
    print('Number of sections per neurite (apical dendrites):',
          nm.get('number_of_sections_per_neurite',
                 nrn, neurite_type=nm.NeuriteType.apical_dendrite))
    print('Number of sections per neurite (basal dendrites):',
          nm.get('number_of_sections_per_neurite',
                 nrn, neurite_type=nm.NeuriteType.apical_dendrite))

    # OK, this is getting repetitive, so lets loop over valid neurite types.
    # The following methods return arrays of measurements. We will gather some
    # summary statistics for each and print them.

    # Section lengths for all and different types of neurite
    for ttype in nm.NEURITE_TYPES:
        sec_len = nm.get('section_lengths', nrn, neurite_type=ttype)
        print('Section lengths (', ttype, '):', sep='')
        pprint_stats(sec_len)

    # Segment lengths for all and different types of neurite
    for ttype in nm.NEURITE_TYPES:
        seg_len = nm.get('segment_lengths', nrn, neurite_type=ttype)
        print('Segment lengths (', ttype, '):', sep='')
        pprint_stats(seg_len)

    # Section radial distances for all and different types of neurite
    # Careful! Here we need to pass tree type as a named argument
    for ttype in nm.NEURITE_TYPES:
        sec_rad_dist = nm.get('section_radial_distances', nrn, neurite_type=ttype)
        print('Section radial distance (', ttype, '):', sep='')
        pprint_stats(sec_rad_dist)

    # Section path distances for all and different types of neurite
    # Careful! Here we need to pass tree type as a named argument
    for ttype in nm.NEURITE_TYPES:
        sec_path_dist = nm.get('section_path_distances', nrn, neurite_type=ttype)
        print('Section path distance (', ttype, '):', sep='')
        pprint_stats(sec_path_dist)

    # Local bifurcation angles for all and different types of neurite
    for ttype in nm.NEURITE_TYPES:
        local_bifangles = nm.get('local_bifurcation_angles', nrn, neurite_type=ttype)
        print('Local bifurcation angles (', ttype, '):', sep='')
        pprint_stats(local_bifangles)

    # Remote bifurcation angles for all and different types of neurite
    for ttype in nm.NEURITE_TYPES:
        rem_bifangles = nm.get('remote_bifurcation_angles', nrn, neurite_type=ttype)
        print('Local bifurcation angles (', ttype, '):', sep='')
        pprint_stats(rem_bifangles)



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Neurite functions."""

from itertools import chain

import numpy as np

from neurom import morphmath
from neurom.core import Tree, iter_neurites, iter_sections, iter_segments, NeuriteType
from neurom.core.dataformat import COLS
from neurom.core.types import tree_type_checker as is_type
from neurom.features import bifurcationfunc, neuronfunc, sectionfunc
from neurom.features.sectionfunc import downstream_pathlength
from neurom.features.bifurcationfunc import partition_asymmetry
from neurom.geom import convex_hull
from neurom.morphmath import interval_lengths


def total_length(nrn_pop, neurite_type=NeuriteType.all):
    """Get the total length of all sections in the group of neurons or neurites."""
    nrns = neuronfunc.neuron_population(nrn_pop)
    return list(sum(section_lengths(n, neurite_type=neurite_type)) for n in nrns)


def n_segments(neurites, neurite_type=NeuriteType.all):
    """Number of segments in a collection of neurites."""
    return sum(len(s.points) - 1
               for s in iter_sections(neurites, neurite_filter=is_type(neurite_type)))


def n_neurites(neurites, neurite_type=NeuriteType.all):
    """Number of neurites in a collection of neurites."""
    return sum(1 for _ in iter_neurites(neurites, filt=is_type(neurite_type)))


def n_sections(neurites, neurite_type=NeuriteType.all, iterator_type=Tree.ipreorder):
    """Number of sections in a collection of neurites."""
    return sum(1 for _ in iter_sections(neurites,
                                        iterator_type=iterator_type,
                                        neurite_filter=is_type(neurite_type)))


def n_bifurcation_points(neurites, neurite_type=NeuriteType.all):
    """Number of bifurcation points in a collection of neurites."""
    return n_sections(neurites, neurite_type=neurite_type, iterator_type=Tree.ibifurcation_point)


def n_forking_points(neurites, neurite_type=NeuriteType.all):
    """Number of forking points in a collection of neurites."""
    return n_sections(neurites, neurite_type=neurite_type, iterator_type=Tree.iforking_point)


def n_leaves(neurites, neurite_type=NeuriteType.all):
    """Number of leaves points in a collection of neurites."""
    return n_sections(neurites, neurite_type=neurite_type, iterator_type=Tree.ileaf)


def total_area_per_neurite(neurites, neurite_type=NeuriteType.all):
    """Surface area in a collection of neurites.

    The area is defined as the sum of the area of the sections.
    """
    return [neurite.area for neurite in iter_neurites(neurites, filt=is_type(neurite_type))]


def map_sections(fun, neurites, neurite_type=NeuriteType.all, iterator_type=Tree.ipreorder):
    """Map `fun` to all the sections in a collection of neurites."""
    return map(fun, iter_sections(neurites,
                                  iterator_type=iterator_type,
                                  neurite_filter=is_type(neurite_type)))


def _section_length(section):
    """Get section length of `section`."""
    return morphmath.section_length(section.points)


def section_lengths(neurites, neurite_type=NeuriteType.all):
    """Section lengths in a collection of neurites."""
    return map_sections(_section_length, neurites, neurite_type=neurite_type)


def section_term_lengths(neurites, neurite_type=NeuriteType.all):
    """Termination section lengths in a collection of neurites."""
    return map_sections(_section_length, neurites, neurite_type=neurite_type,
                        iterator_type=Tree.ileaf)


def section_bif_lengths(neurites, neurite_type=NeuriteType.all):
    """Bifurcation section lengths in a collection of neurites."""
    return map_sections(_section_length, neurites, neurite_type=neurite_type,
                        iterator_type=Tree.ibifurcation_point)


def section_branch_orders(neurites, neurite_type=NeuriteType.all):
    """Section branch orders in a collection of neurites."""
    return map_sections(sectionfunc.branch_order, neurites, neurite_type=neurite_type)


def section_bif_branch_orders(neurites, neurite_type=NeuriteType.all):
    """Bifurcation section branch orders in a collection of neurites."""
    return map_sections(sectionfunc.branch_order, neurites, neurite_type=neurite_type,
                        iterator_type=Tree.ibifurcation_point)


def section_term_branch_orders(neurites, neurite_type=NeuriteType.all):
    """Termination section branch orders in a collection of neurites."""
    return map_sections(sectionfunc.branch_order, neurites, neurite_type=neurite_type,
                        iterator_type=Tree.ileaf)


def section_path_lengths(neurites, neurite_type=NeuriteType.all):
    """Path lengths of a collection of neurites."""
    # Calculates and stores the section lengths in one pass,
    # then queries the lengths in the path length iterations.
    # This avoids repeatedly calculating the lengths of the
    # same sections.
    dist = {}
    neurite_filter = is_type(neurite_type)

    for s in iter_sections(neurites, neurite_filter=neurite_filter):
        dist[s] = s.length

    def pl2(node):
        """Calculate the path length using cached section lengths."""
        return sum(dist[n] for n in node.iupstream())

    return map_sections(pl2, neurites, neurite_type=neurite_type)


def map_neurons(fun, neurites, neurite_type):
    """Map `fun` to all the neurites in a single or collection of neurons."""
    nrns = neuronfunc.neuron_population(neurites)
    return [fun(n, neurite_type=neurite_type) for n in nrns]


def number_of_sections(neurites, neurite_type=NeuriteType.all):
    """Number of sections in a collection of neurites."""
    return map_neurons(n_sections, neurites, neurite_type)


def number_of_neurites(neurites, neurite_type=NeuriteType.all):
    """Number of neurites in a collection of neurites."""
    return map_neurons(n_neurites, neurites, neurite_type)


def number_of_bifurcations(neurites, neurite_type=NeuriteType.all):
    """Number of bifurcation points in a collection of neurites."""
    return map_neurons(n_bifurcation_points, neurites, neurite_type)


def number_of_forking_points(neurites, neurite_type=NeuriteType.all):
    """Number of forking points in a collection of neurites."""
    return map_neurons(n_forking_points, neurites, neurite_type)


def number_of_terminations(neurites, neurite_type=NeuriteType.all):
    """Number of leaves points in a collection of neurites."""
    return map_neurons(n_leaves, neurites, neurite_type)


def number_of_segments(neurites, neurite_type=NeuriteType.all):
    """Number of sections in a collection of neurites."""
    return map_neurons(n_segments, neurites, neurite_type)


def map_segments(func, neurites, neurite_type):
    """Map `func` to all the segments in a collection of neurites.

    `func` accepts a section and returns list of values corresponding to each segment.
    """
    neurite_filter = is_type(neurite_type)
    return [
        s for ss in iter_sections(neurites, neurite_filter=neurite_filter) for s in func(ss)
    ]


def segment_lengths(neurites, neurite_type=NeuriteType.all):
    """Lengths of the segments in a collection of neurites."""
    return map_segments(sectionfunc.segment_lengths, neurites, neurite_type)


def segment_areas(neurites, neurite_type=NeuriteType.all):
    """Areas of the segments in a collection of neurites."""
    return [morphmath.segment_area(seg) for seg
            in iter_segments(neurites, is_type(neurite_type))]


def segment_volumes(neurites, neurite_type=NeuriteType.all):
    """Volumes of the segments in a collection of neurites."""
    def _func(sec):
        """List of segment volumes of a section."""
        return [morphmath.segment_volume(seg) for seg in zip(sec.points[:-1], sec.points[1:])]

    return map_segments(_func, neurites, neurite_type)


def segment_radii(neurites, neurite_type=NeuriteType.all):
    """Arithmetic mean of the radii of the points in segments in a collection of neurites."""
    def _seg_radii(sec):
        """Vectorized mean radii."""
        pts = sec.points[:, COLS.R]
        return np.divide(np.add(pts[:-1], pts[1:]), 2.0)

    return map_segments(_seg_radii, neurites, neurite_type)


def segment_taper_rates(neurites, neurite_type=NeuriteType.all):
    """Diameters taper rates of the segments in a collection of neurites.

    The taper rate is defined as the absolute radii differences divided by length of the section
    """
    def _seg_taper_rates(sec):
        """Vectorized taper rates."""
        pts = sec.points[:, COLS.XYZR]
        diff = np.diff(pts, axis=0)
        distance = np.linalg.norm(diff[:, COLS.XYZ], axis=1)
        return np.divide(2 * np.abs(diff[:, COLS.R]), distance)

    return map_segments(_seg_taper_rates, neurites, neurite_type)


def section_taper_rates(neurites, neurite_type=NeuriteType.all):
    """Diameter taper rates of the sections in a collection of neurites from root to tip.

    Taper rate is defined here as the linear fit along a section.
    It is expected to be negative for neurons.
    """
    def _sec_taper_rate(sec):
        """Taper rate from fit along a section."""
        path_distances = np.cumsum(interval_lengths(sec.points, prepend_zero=True))
        return np.polynomial.polynomial.polyfit(path_distances, 2 * sec.points[:, COLS.R], 1)[1]

    return map_sections(_sec_taper_rate, neurites, neurite_type=neurite_type)


def segment_meander_angles(neurites, neurite_type=NeuriteType.all):
    """Inter-segment opening angles in a section."""
    return list(chain.from_iterable(map_sections(
        sectionfunc.section_meander_angles, neurites, neurite_type)))


def segment_midpoints(neurites, neurite_type=NeuriteType.all):
    """Return a list of segment mid-points in a collection of neurites."""
    def _seg_midpoint(sec):
        """Return the mid-points of segments in a section."""
        pts = sec.points[:, COLS.XYZ]
        return np.divide(np.add(pts[:-1], pts[1:]), 2.0)

    return map_segments(_seg_midpoint, neurites, neurite_type)


def segment_path_lengths(neurites, neurite_type=NeuriteType.all):
    """Returns pathlengths between all non-root points and their root point."""
    pathlength = {}
    neurite_filter = is_type(neurite_type)

    def _get_pathlength(section):
        if section.id not in pathlength:
            if section.parent:
                pathlength[section.id] = section.parent.length + _get_pathlength(section.parent)
            else:
                pathlength[section.id] = 0
        return pathlength[section.id]

    result = [_get_pathlength(section) + np.cumsum(sectionfunc.segment_lengths(section))
              for section in iter_sections(neurites, neurite_filter=neurite_filter)]
    return np.hstack(result) if result else np.array([])


def segment_radial_distances(neurites, neurite_type=NeuriteType.all, origin=None):
    """Returns the list of distances between all segment mid points and origin."""
    def _radial_distances(sec, pos):
        """List of distances between the mid point of each segment and pos."""
        mid_pts = 0.5 * (sec.points[:-1, COLS.XYZ] + sec.points[1:, COLS.XYZ])
        return np.linalg.norm(mid_pts - pos[COLS.XYZ], axis=1)

    dist = []
    for n in iter_neurites(neurites, filt=is_type(neurite_type)):
        pos = n.root_node.points[0] if origin is None else origin
        dist.extend([s for ss in n.iter_sections() for s in _radial_distances(ss, pos)])

    return dist


def local_bifurcation_angles(neurites, neurite_type=NeuriteType.all):
    """Get a list of local bifurcation angles in a collection of neurites."""
    return map_sections(bifurcationfunc.local_bifurcation_angle,
                        neurites,
                        neurite_type=neurite_type,
                        iterator_type=Tree.ibifurcation_point)


def remote_bifurcation_angles(neurites, neurite_type=NeuriteType.all):
    """Get a list of remote bifurcation angles in a collection of neurites."""
    return map_sections(bifurcationfunc.remote_bifurcation_angle,
                        neurites,
                        neurite_type=neurite_type,
                        iterator_type=Tree.ibifurcation_point)


def bifurcation_partitions(neurites, neurite_type=NeuriteType.all):
    """Partition at bifurcation points of a collection of neurites."""
    return map(bifurcationfunc.bifurcation_partition,
               iter_sections(neurites,
                             iterator_type=Tree.ibifurcation_point,
                             neurite_filter=is_type(neurite_type)))


def partition_asymmetries(neurites, neurite_type=NeuriteType.all, variant='branch-order'):
    """Partition asymmetry at bifurcation points of a collection of neurites.

    Variant: length is a different definition, as the absolute difference in
    downstream path lenghts, relative to the total neurite path length
    """
    if variant not in {'branch-order', 'length'}:
        raise ValueError('Please provide a valid variant for partition asymmetry,\
                         found %s' % variant)

    if variant == 'branch-order':
        return map(partition_asymmetry,
                   iter_sections(neurites,
                                 iterator_type=Tree.ibifurcation_point,
                                 neurite_filter=is_type(neurite_type)))

    asymmetries = list()
    for neurite in iter_neurites(neurites, filt=is_type(neurite_type)):
        neurite_length = total_length_per_neurite(neurite)[0]
        for section in iter_sections(neurite,
                                     iterator_type=Tree.ibifurcation_point,
                                     neurite_filter=is_type(neurite_type)):
            pathlength_diff = abs(downstream_pathlength(section.children[0]) -
                                  downstream_pathlength(section.children[1]))
            asymmetries.append(pathlength_diff / neurite_length)
    return asymmetries


def sibling_ratios(neurites, neurite_type=NeuriteType.all, method='first'):
    """Sibling ratios at bifurcation points of a collection of neurites.

    The sibling ratio is the ratio between the diameters of the
    smallest and the largest child. It is a real number between
    0 and 1. Method argument allows one to consider mean diameters
    along the child section instead of diameter of the first point.
    """
    return map(lambda bif_point: bifurcationfunc.sibling_ratio(bif_point, method),
               iter_sections(neurites,
                             iterator_type=Tree.ibifurcation_point,
                             neurite_filter=is_type(neurite_type)))


def partition_pairs(neurites, neurite_type=NeuriteType.all):
    """Partition pairs at bifurcation points of a collection of neurites.

    Partition pair is defined as the number of bifurcations at the two
    daughters of the bifurcating section
    """
    return map(bifurcationfunc.partition_pair,
               iter_sections(neurites,
                             iterator_type=Tree.ibifurcation_point,
                             neurite_filter=is_type(neurite_type)))


def diameter_power_relations(neurites, neurite_type=NeuriteType.all, method='first'):
    """Calculate the diameter power relation at a bifurcation point.

    Diameter power relation is defined in https://www.ncbi.nlm.nih.gov/pubmed/18568015

    This quantity gives an indication of how far the branching is from
    the Rall ratio (when =1).
    """
    return (bifurcationfunc.diameter_power_relation(bif_point, method)
            for bif_point in iter_sections(neurites,
                                           iterator_type=Tree.ibifurcation_point,
                                           neurite_filter=is_type(neurite_type)))


def section_radial_distances(neurites, neurite_type=NeuriteType.all, origin=None,
                             iterator_type=Tree.ipreorder):
    """Section radial distances in a collection of neurites.

    The iterator_type can be used to select only terminal sections (ileaf)
    or only bifurcations (ibifurcation_point).
    """
    dist = []
    for n in iter_neurites(neurites, filt=is_type(neurite_type)):
        pos = n.root_node.points[0] if origin is None else origin
        dist.extend(sectionfunc.section_radial_distance(s, pos)
                    for s in iter_sections(n,
                                           iterator_type=iterator_type))
    return dist


def section_term_radial_distances(neurites, neurite_type=NeuriteType.all, origin=None):
    """Get the radial distances of the termination sections for a collection of neurites."""
    return section_radial_distances(neurites, neurite_type=neurite_type, origin=origin,
                                    iterator_type=Tree.ileaf)


def section_bif_radial_distances(neurites, neurite_type=NeuriteType.all, origin=None):
    """Get the radial distances of the bifurcation sections for a collection of neurites."""
    return section_radial_distances(neurites, neurite_type=neurite_type, origin=origin,
                                    iterator_type=Tree.ibifurcation_point)


def number_of_sections_per_neurite(neurites, neurite_type=NeuriteType.all):
    """Get the number of sections per neurite in a collection of neurites."""
    return list(sum(1 for _ in n.iter_sections())
                for n in iter_neurites(neurites, filt=is_type(neurite_type)))


def total_length_per_neurite(neurites, neurite_type=NeuriteType.all):
    """Get the path length per neurite in a collection."""
    return list(sum(s.length for s in n.iter_sections())
                for n in iter_neurites(neurites, filt=is_type(neurite_type)))


def terminal_path_lengths_per_neurite(neurites, neurite_type=NeuriteType.all):
    """Get the path lengths to each terminal point per neurite in a collection."""
    return list(sectionfunc.section_path_length(s)
                for n in iter_neurites(neurites, filt=is_type(neurite_type))
                for s in iter_sections(n, iterator_type=Tree.ileaf))


def total_volume_per_neurite(neurites, neurite_type=NeuriteType.all):
    """Get the volume per neurite in a collection."""
    return list(sum(s.volume for s in n.iter_sections())
                for n in iter_neurites(neurites, filt=is_type(neurite_type)))


def neurite_volume_density(neurites, neurite_type=NeuriteType.all):
    """Get the volume density per neurite.

    The volume density is defined as the ratio of the neurite volume and
    the volume of the neurite's enclosing convex hull
    """
    def vol_density(neurite):
        """Volume density of a single neurite."""
        return neurite.volume / convex_hull(neurite).volume

    return list(vol_density(n)
                for n in iter_neurites(neurites, filt=is_type(neurite_type)))


def section_volumes(neurites, neurite_type=NeuriteType.all):
    """Section volumes in a collection of neurites."""
    return map_sections(sectionfunc.section_volume, neurites, neurite_type=neurite_type)


def section_areas(neurites, neurite_type=NeuriteType.all):
    """Section areas in a collection of neurites."""
    return map_sections(sectionfunc.section_area, neurites, neurite_type=neurite_type)


def section_tortuosity(neurites, neurite_type=NeuriteType.all):
    """Section tortuosities in a collection of neurites."""
    return map_sections(sectionfunc.section_tortuosity, neurites, neurite_type=neurite_type)


def section_end_distances(neurites, neurite_type=NeuriteType.all):
    """Section end to end distances in a collection of neurites."""
    return map_sections(sectionfunc.section_end_distance, neurites, neurite_type=neurite_type)


def principal_direction_extents(neurites, neurite_type=NeuriteType.all, direction=0):
    """Principal direction extent of neurites in neurons."""
    def _pde(neurite):
        """Get the PDE of a single neurite."""
        # Get the X, Y,Z coordinates of the points in each section
        points = neurite.points[:, :3]
        return morphmath.principal_direction_extent(points)[direction]

    return [_pde(neurite) for neurite in iter_neurites(neurites, filt=is_type(neurite_type))]


def section_strahler_orders(neurites, neurite_type=NeuriteType.all):
    """Inter-segment opening angles in a section."""
    return map_sections(sectionfunc.strahler_order, neurites, neurite_type)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from copy import deepcopy
from io import StringIO
from pathlib import Path

import numpy as np
from neurom import load_neuron
from neurom.check import morphtree as mt
from neurom.core import Neurite, NeuriteType, Section
from neurom.core.dataformat import COLS
from nose import tools as nt

DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data'
SWC_PATH = DATA_PATH / 'swc'


def _make_flat(neuron):

    class Flattenizer(object):
        def __call__(self, points):
            points = deepcopy(points)
            points[:, COLS.Z] = 0.;
            return points

    return neuron.transform(Flattenizer())


def _make_monotonic(neuron):
    for neurite in neuron.neurites:
        for node in neurite.iter_sections():
            sec = node.points
            if node.parent is not None:
                sec[0][COLS.R] = node.parent.points[-1][COLS.R] / 2.
            for point_id in range(len(sec) - 1):
                sec[point_id + 1][COLS.R] = sec[point_id][COLS.R] / 2.


def _generate_neurite(mode):

    def new_radius(prev_radius, mode):
        if mode == 0:
            return prev_radius/2.
        elif mode == 1:
            return prev_radius
        else:
            return prev_radius * 2.

    def fake_tree(init_radius, mode):

        radius = init_radius

        sec = []
        for _ in range(5):
            sec.append([0, 0, 0, radius, 0, 0])
            radius = new_radius(radius, mode)

        return Section(np.array(sec))

    radius = 1.

    tree = fake_tree(radius, mode)
    tree.add_child(fake_tree(tree.points[-1][COLS.R], mode))
    tree.add_child(fake_tree(tree.points[-1][COLS.R], mode))
    tree.type = NeuriteType.undefined

    return Neurite(tree)


def _generate_back_track_tree(n, dev):
    points = np.array(dev) + np.array([1, 3 if n == 0 else -3, 0])

    neuron = load_neuron(StringIO(u"""
    ((CellBody)
     (0 0 0 0.4))

    ((Dendrite)
    (0 0 0 0.4)
    (0 1 0 0.3)
    (0 2 0 0.28)
    (
      (0 2 0 0.28)
      (1 3 0 0.3)
      (2 4 0 0.22)
      |
      (0 2 0 0.28)
      (1 -3 0 0.3)
      (2 -4 0 0.24)
      ({0} {1} {2} 0.52)
      (3 -5 0 0.2)
      (4 -6 0 0.2)
    ))
    """.format(*points.tolist())), reader='asc')

    return neuron.neurites[0]


# def test_is_flat():

#     neu_tree = load_neuron(Path(SWC_PATH, 'Neuron.swc')).neurites[0]

#     nt.assert_false(mt.is_flat(neu_tree, 1e-6, method='tolerance'))
#     nt.assert_false(mt.is_flat(neu_tree, 0.1, method='ratio'))


# FIXME: this test fails because of https://github.com/BlueBrain/MorphIO/issues/29
@nt.nottest
def test_is_back_tracking():
    # case 1: a back-track falls directly on a previous node
    t = _generate_back_track_tree(1, (0., 0., 0.))
    nt.assert_true(mt.is_back_tracking(t))

    # case 2: a zigzag is close to another segment
    t = _generate_back_track_tree(1, (0.1, -0.1, 0.02))
    nt.assert_true(mt.is_back_tracking(t))

    # case 3: a zigzag is close to another segment 2
    t = _generate_back_track_tree(1, (-0.2, 0.04, 0.144))
    nt.assert_true(mt.is_back_tracking(t))

    # case 4: a zigzag far from civilization
    t = _generate_back_track_tree(1, (10., -10., 10.))
    nt.assert_false(mt.is_back_tracking(t))

    # case 5: a zigzag on another section
    # currently zigzag is defined on the same section
    # thus this test should not be true
    t = _generate_back_track_tree(0, (-0.2, 0.04, 0.144))
    nt.assert_false(mt.is_back_tracking(t))


def test_get_flat_neurites():

    n = load_neuron(Path(SWC_PATH, 'Neuron.swc'))

    nt.assert_equal(len(mt.get_flat_neurites(n, 1e-6, method='tolerance')), 0)
    nt.assert_equal(len(mt.get_flat_neurites(n, 0.1, method='ratio')), 0)

    n = _make_flat(n)

    nt.assert_equal(len(mt.get_flat_neurites(n, 1e-6, method='tolerance')), 4)
    nt.assert_equal(len(mt.get_flat_neurites(n, 0.1, method='ratio')), 4)


def test_get_back_tracking_neurites():

    n = load_neuron(Path(SWC_PATH, 'Neuron.swc'))
    nt.assert_equal(len(mt.get_back_tracking_neurites(n)), 4)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

import numpy as np
from nose import tools as nt

import neurom as nm
from neurom import geom

SWC_DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data/swc'
NRN = nm.load_neuron(Path(SWC_DATA_PATH, 'Neuron.swc'))
SIMPLE = nm.load_neuron(SWC_DATA_PATH / 'simple.swc')


class PointObj(object):
    pass


def test_bounding_box():

    pts = np.array([[-1, -2, -3, -999],
                    [1, 2, 3, 1000],
                    [-100, 5, 33, 42],
                    [42, 55, 12, -3]])

    obj = PointObj()
    obj.points = pts

    nt.assert_true(np.alltrue(geom.bounding_box(obj) == [[-100, -2, -3], [42, 55, 33]]))


def test_bounding_box_neuron():

    ref = np.array([[-40.32853516, -57.600172, 0.],
                    [64.74726272, 48.51626225, 54.20408797]])

    nt.assert_true(np.allclose(geom.bounding_box(NRN), ref))


def test_bounding_box_soma():
    ref = np.array([[0., 0., 0.], [0.1, 0.2, 0.]])
    nt.assert_true(np.allclose(geom.bounding_box(NRN.soma), ref))


def test_bounding_box_neurite():
    nrt = SIMPLE.neurites[0]
    ref = np.array([[-5.,  0.,  0.], [ 6.,  5.,  0.]])
    np.testing.assert_allclose(geom.bounding_box(nrt), ref)


def test_convex_hull_points():

    # This leverages scipy ConvexHull and we don't want
    # to re-test scipy, so simply check that the points are the same.
    hull = geom.convex_hull(NRN)
    nt.ok_(np.alltrue(hull.points == NRN.points[:, :3]))


def test_convex_hull_volume():

    # This leverages scipy ConvexHull and we don't want
    # to re-test scipy, so simply regression test the volume
    hull = geom.convex_hull(NRN)
    nt.assert_almost_equal(hull.volume, 208639.7805986741, places=2)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from pathlib import Path

from neurom import load_neuron
from neurom.check import structural_checks as chk
from neurom.exceptions import MissingParentError, SomaError
from nose import tools as nt

DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data'
SWC_PATH = Path(DATA_PATH, 'swc')
H5V1_PATH = Path(DATA_PATH, 'h5/v1')

# TODO:
# The origin NeuronM neuron loader was more flexible than the one in MorphIO
# As a result it was possible to load broken morphologies and call check function on them
# to see if they are broken or nothing
# The current implementation of MorphIO does not support this but it would be a good nice-to-have
# in the future

class TestIOCheckFST(object):
    def setup(self):
        self.load_neuron = load_neuron

    def test_everything_is_fine(self):

        files = [Path(SWC_PATH, f)
                 for f in ['Neuron.swc',
                           'Single_apical_no_soma.swc',
                           'Single_apical.swc',
                           'Single_basal.swc',
                           'Single_axon.swc',
                           'Neuron_zero_radius.swc',
                           'sequential_trunk_off_0_16pt.swc',
                           'sequential_trunk_off_1_16pt.swc',
                           'sequential_trunk_off_42_16pt.swc',
                           'Neuron_no_missing_ids_no_zero_segs.swc']
                 ]

        for f in files:
            nt.ok_(self.load_neuron(f))



    @nt.raises(MissingParentError)
    def test_has_sequential_ids_bad_data(self):
        nt.ok_(self.load_neuron(os.path.join(SWC_PATH, 'Neuron_missing_ids.swc')))

    def test_is_single_tree_bad_data(self):
        nt.ok_(self.load_neuron(os.path.join(SWC_PATH, 'Neuron_disconnected_components.swc')))

    @nt.raises(SomaError)
    def test_multiple_somata(self):
        nt.ok_(self.load_neuron(os.path.join(SWC_PATH, 'multiple_somata.swc')))

    @nt.raises(MissingParentError)
    def test_has_no_missing_parents_bad_data(self):
        nt.ok_(self.load_neuron(os.path.join(SWC_PATH, 'Neuron_missing_parents.swc')))

    def test_has_soma_points_bad_data(self):
        f = os.path.join(SWC_PATH, 'Single_apical_no_soma.swc')
        nt.ok_(not chk.has_soma_points(self.load_neuron(f)))


    # @nt.raises(MissingParentError)
    def test_has_valid_soma_bad_data(self):
        dw = self.load_neuron(os.path.join(SWC_PATH, 'Single_apical_no_soma.swc'))
        # nt.ok_(not chk.has_valid_soma(dw))

    @nt.raises(NotImplementedError)
    def test_has_finite_radius_neurites_good_data(self):
        files = [Path(SWC_PATH, f)
                 for f in ['Neuron.swc',
                           'Single_apical.swc',
                           'Single_basal.swc',
                           'Single_axon.swc']]

        files.append(Path(H5V1_PATH, 'Neuron_2_branch.h5'))

        for f in files:
            ok = chk.has_all_finite_radius_neurites(self.load_neuron(f))
            nt.ok_(ok)
            nt.ok_(len(ok.info) == 0)

    @nt.raises(NotImplementedError)
    def test_has_finite_radius_neurites_bad_data(self):
        f = os.path.join(SWC_PATH, 'Neuron_zero_radius.swc')
        ok = chk.has_all_finite_radius_neurites(self.load_neuron(f))
        nt.ok_(not ok)
        nt.ok_(list(ok.info) == [194, 210, 246, 304, 493])



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Soma classes and functions."""
import math
import warnings

import numpy as np

from neurom import morphmath, SomaType
from neurom.core.dataformat import COLS


class Soma(object):
    """Base class for a soma.

    Holds a list of raw data rows corresponding to soma points
    and provides iterator access to them.
    """

    def __init__(self, morphio_soma):
        self._morphio_soma = morphio_soma

    @property
    def center(self):
        '''Obtain the center from the first stored point'''
        return self._morphio_soma.points[0]

    def iter(self):
        '''Iterator to soma contents'''
        return iter(self.points)

    @property
    def points(self):
        '''Get the set of (x, y, z, r) points this soma'''
        return np.concatenate((self._morphio_soma.points,
                               self._morphio_soma.diameters[:, np.newaxis] / 2.),
                              axis=1)

    @points.setter
    def points(self, value):
        '''Set the points'''
        value = np.asarray(value)
        self._morphio_soma.points = np.copy(value[:, COLS.XYZ])
        self._morphio_soma.diameters = np.copy(value[:, COLS.R]) * 2

    @property
    def volume(self):
        """Gets soma volume assuming it is a sphere."""
        warnings.warn('Approximating soma volume by a sphere. {}'.format(self))
        return 4. / 3 * math.pi * self.radius ** 3


class SomaSinglePoint(Soma):
    """Type A: 1point soma.

    Represented by a single point.
    """

    def __init__(self, morphio_soma):
        super(SomaSinglePoint, self).__init__(morphio_soma)
        self.radius = self.points[0][COLS.R]

    def __str__(self):
        """Return a string representation."""
        return ('SomaSinglePoint(%s) <center: %s, radius: %s>' %
                (repr(self.points), self.center, self.radius))


class SomaCylinders(Soma):
    """Soma composed of cylinders (like in SWC).

    points describe the locations of the cylinder start/end points, with
    their respective radii, much like how neurites are described:

        ex::

                             /)
                            / o)
                    ______ /  )
                (|)      ) / /
                ( o )    o )/
                (|)_____ )

      Here we have a 'side-view', with each 'o' representing a point, and the
      radius is the height of a '|' character, and the ')' try and show the
      curvature of the cylinger

    Note: when, as in the case above, the cylinder center points don't lie
    in a line, then the overlap between cylinders isn't taken into account for
    the area calculation
    """

    def __init__(self, morphio_soma):
        super(SomaCylinders, self).__init__(morphio_soma)
        self.area = sum(morphmath.segment_area((p0, p1))
                        for p0, p1 in zip(self.points, self.points[1:]))
        self.radius = math.sqrt(self.area / (4. * math.pi))

    @property
    def center(self):
        """Obtain the center from the first stored point."""
        return self.points[0][COLS.XYZ]

    @property
    def volume(self):
        """Return the volume of soma."""
        return sum(morphmath.segment_volume((p0, p1))
                   for p0, p1 in zip(self.points, self.points[1:]))

    def __str__(self):
        """Return a string representation."""
        return ('SomaCylinders(%s) <center: %s, virtual radius: %s>' %
                (repr(self.points), self.center, self.radius))


class SomaNeuromorphoThreePointCylinders(SomaCylinders):
    """NeuroMorpho compatible soma.

    Reference:
        http://neuromorpho.org/SomaFormat.html

        Quote (Tue Feb 28 14:26:56 CET 2017):
        The first point constitutes the center of the soma, with coordinates (xs, ys,
        zs) corresponding to the average of all the points in the single contour. An
        equivalent radius (rs) is computed as the average distance of each point of the
        single contour from this center. The first point is assigned parent ID -1 as is
        standard for the root in the SWC format. The second and third soma points, as
        well as all starting points (roots) of dendritic and axonal arbors have this
        first point as the parent (parent ID 1). The x and z coordinates as well as the
        radius of both the second and third soma points are the same as that of the
        first point. The y coordinates of the second and third points are shifted by -rs
        and +rs, respectively (Figure 2). The surface area of this soma cylinder equals
        the surface area of a sphere of radius rs.
    """

    def __init__(self, morphio_soma):
        """Initialize a SomaNeuromorphoThreePointCylinders object."""
        super(SomaNeuromorphoThreePointCylinders, self).__init__(morphio_soma)

        # X    Y     Z   R    P
        # xs ys      zs rs   -1
        # xs (ys-rs) zs rs    1
        # xs (ys+rs) zs rs    1

        r = self.points[0, COLS.R]
        # make sure the above invariant holds
        assert (np.isclose(r, self.points[1, COLS.R]) and np.isclose(r, self.points[2, COLS.R])), \
            'All radii must be the same'
        # only warn users about invalid format
        if r < 1e-5:
            warnings.warn('Zero radius for {}'.format(self))
        if not np.isclose(self.points[0, COLS.Y] - self.points[1, COLS.Y], r):
            warnings.warn(
                'The second point must be one radius below 0 on the y-plane for {}'.format(self))
        if not np.isclose(self.points[0, COLS.Y] - self.points[2, COLS.Y], -r):
            warnings.warn(
                'The third point must be one radius above 0 on the y-plane for {}'.format(self))
        h = morphmath.point_dist(self.points[1, COLS.XYZ], self.points[2, COLS.XYZ])
        self.area = 2.0 * math.pi * r * h  # ignores the 'end-caps' of the cylinder
        self.radius = math.sqrt(self.area / (4. * math.pi))

    @property
    def volume(self):
        """Return the volume of the soma."""
        return 2 * math.pi * self.radius ** 3

    def __str__(self):
        """Return a string representation."""
        return ('SomaNeuromorphoThreePointCylinders(%s) <center: %s, radius: %s>' %
                (repr(self.points), self.center, self.radius))


class SomaSimpleContour(Soma):
    """Type C: multiple points soma.

    Represented by a contour.

    The equivalent radius as the average distance to the center.

    Note: This doesn't currently check to see if the contour is in a plane. Also
    the radii of the points are not taken into account.
    """

    def __init__(self, morphio_soma):
        """Initialize a SomaSimpleContour object."""
        super(SomaSimpleContour, self).__init__(morphio_soma)
        self.radius = morphmath.average_points_dist(
            self.center, self.points[:, COLS.XYZ])

    @property
    def center(self):
        """Obtain the center from the average of all points."""
        return np.mean(self.points[:, COLS.XYZ], axis=0)

    def __str__(self):
        """Return a string representation."""
        return ('SomaSimpleContour(%s) <center: %s, radius: %s>' %
                (repr(self.points), self.center, self.radius))


def make_soma(morphio_soma):
    """Make a soma object from a set of points

    Parameters:
        soma_type: the type of soma
        points: collection of points forming a soma.
    """
    SomaBuilders = {
        SomaType.SOMA_SINGLE_POINT: SomaSinglePoint,
        SomaType.SOMA_CYLINDERS: SomaCylinders,
        SomaType.SOMA_NEUROMORPHO_THREE_POINT_CYLINDERS: SomaNeuromorphoThreePointCylinders,
        SomaType.SOMA_SIMPLE_CONTOUR: SomaSimpleContour,
    }

    return SomaBuilders.get(morphio_soma.type, SomaSimpleContour)(morphio_soma)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from copy import deepcopy

from pathlib import Path
import numpy as np
from numpy.testing import assert_array_equal
from nose import tools as nt

import neurom as nm
from neurom.core import iter_segments, iter_neurites, graft_neuron, Neuron
from morphio import Morphology as ImmutMorphology
from morphio.mut import Morphology


SWC_PATH = Path(__file__).parent.parent.parent.parent / 'test_data/swc/'


def test_simple():
    nrn1 = nm.load_neuron(Path(SWC_PATH, 'simple.swc'))

def test_load_neuron_pathlib():
    nrn1 = nm.load_neuron(Path(SWC_PATH, 'simple.swc'))

def test_load_neuron_from_other_neurons():
    filename = Path(SWC_PATH, 'simple.swc')

    expected_points = [[ 0.,  0.,  0.,  1.],
                       [ 0.,  5.,  0.,  1.],
                       [ 0.,  5.,  0.,  1.],
                       [-5.,  5.,  0.,  0.],
                       [ 0.,  5.,  0.,  1.],
                       [ 6.,  5.,  0.,  0.],
                       [ 0.,  0.,  0.,  1.],
                       [ 0., -4.,  0.,  1.],
                       [ 0., -4.,  0.,  1.],
                       [ 6., -4.,  0.,  0.],
                       [ 0., -4.,  0.,  1.],
                       [-5., -4.,  0.,  0.]]

    assert_array_equal(nm.load_neuron(nm.load_neuron(filename)).points,
                       expected_points)

    assert_array_equal(nm.load_neuron(Morphology(filename)).points,
                       expected_points)

    assert_array_equal(nm.load_neuron(ImmutMorphology(filename)).points,
                       expected_points)

def test_for_morphio():
    Neuron(Morphology())

    neuron = Morphology()
    neuron.soma.points = [[0,0,0], [1,1,1], [2,2,2]]
    neuron.soma.diameters = [1, 1, 1]

    NeuroM_neuron = Neuron(neuron)
    assert_array_equal(NeuroM_neuron.soma.points,
                       [[0., 0., 0., 0.5],
                        [1., 1., 1., 0.5],
                        [2., 2., 2., 0.5]])

    NeuroM_neuron.soma.points = [[1, 1, 1, 1],
                                 [2, 2, 2, 2]]
    assert_array_equal(NeuroM_neuron.soma.points,
                       [[1, 1, 1, 1],
                        [2, 2, 2, 2]])


# def test_deep_copy():
#     nrn1 = nm.load_neuron(Path(SWC_PATH, 'simple.swc'))
#     nrn2 = deepcopy(nrn1)
#     check_cloned_neuron(nrn1, nrn2)


def test_graft_neuron():
    nrn1 = nm.load_neuron(Path(SWC_PATH, 'simple.swc'))
    basal_dendrite = nrn1.neurites[0]
    nrn2 = graft_neuron(basal_dendrite.root_node)
    nt.assert_equal(len(nrn2.neurites), 1)
    nt.assert_equal(basal_dendrite, nrn2.neurites[0])


def check_cloned_neuron(nrn1, nrn2):
    # check if two neurons are identical

    # soma
    nt.ok_(isinstance(nrn2.soma, type(nrn1.soma)))
    nt.eq_(nrn1.soma.radius, nrn2.soma.radius)

    for v1, v2 in zip(nrn1.soma.iter(), nrn2.soma.iter()):
        nt.ok_(np.allclose(v1, v2))

    # neurites
    for v1, v2 in zip(iter_segments(nrn1), iter_segments(nrn2)):
        (v1_start, v1_end), (v2_start, v2_end) = v1, v2
        nt.ok_(np.allclose(v1_start, v2_start))
        nt.ok_(np.allclose(v1_end, v2_end))

    # check if the ids are different
    # somata
    nt.ok_(nrn1.soma is not nrn2.soma)

    # neurites
    for neu1, neu2 in zip(nrn1.neurites, nrn2.neurites):
        nt.ok_(neu1 is not neu2)

    # check if changes are propagated between neurons
    nrn2.soma.radius = 10.
    nt.ok_(nrn1.soma.radius != nrn2.soma.radius)


def test_str():
    n = nm.load_neuron(Path(SWC_PATH, 'simple.swc'))
    nt.ok_('Neuron' in str(n))
    nt.ok_('Section' in str(n.neurites[0].root_node))



import textwrap
import warnings
from io import StringIO
from pathlib import Path
import numpy as np
from mock import patch
from nose.tools import eq_, ok_, assert_raises, assert_equal

import neurom as nm
from neurom import load_neuron
import neurom.io as io
from neurom.core.dataformat import COLS
from numpy.testing import assert_array_equal
from nose import tools as nt

from neurom.exceptions import RawDataError, SomaError

DATA_PATH = Path(Path(__file__).parent, '../../../test_data')
NEUROLUCIDA_PATH = Path(DATA_PATH, 'neurolucida')


def test_soma():
    string_section = u'''
                         ("CellBody"
                         (Color Red)
                         (CellBody)
                         (1 1 0 1 S1)
                         (-1 1 0 1 S2)
                         (-1 -1 0 2 S3)
                         )
        '''
    n = nm.load_neuron(string_section, reader='asc')

    assert_array_equal(n.soma.points,
                       [[1, 1, 0, 0.5],
                        [-1, 1, 0, 0.5],
                        [-1, -1, 0, 1]])

    nt.assert_equal(len(n.neurites), 0)


def test_unknown_token():
    with assert_raises(RawDataError) as obj:
        string_section = u'''
                         ("CellBody"
                         (Color Red)
                         (CellBody)
                         (1 1 0 1 S1)
                         (Z 1 0 1 S2) ; <-- Z is a BAD token
                         (-1 -1 0 2 S3)
                         )
        '''
        n = nm.load_neuron(string_section, reader='asc')

    ok_("Unexpected token: Z" in str(obj.exception))
    ok_(":6:error" in str(obj.exception))


def test_unfinished_point():
    with assert_raises(RawDataError) as obj:
        string_section = u'''("CellBody"
                         (Color Red)
                         (CellBody)
                         (1 1'''
        n = nm.load_neuron(string_section, reader='asc')

    ok_('Error converting: "" to float' in str(obj.exception))
    ok_(':4:error' in str(obj.exception))


def test_multiple_soma():
    with assert_raises(SomaError) as obj:
        string_section = u'''
                             ("CellBody"
                             (Color Red)
                             (CellBody)
                             (1 1 0 1 S1)
                             (-1 1 0 1 S2)
                             (-1 -1 0 2 S3)
                             )

                            ("CellBody"
                             (Color Red)
                             (CellBody)
                             (1 1 0 1 S1)
                             (-1 1 0 1 S2)
                             (-1 -1 0 2 S3)
                             )
            '''
        load_neuron(string_section, reader='asc')
    ok_("A soma is already defined" in str(obj.exception))
    ok_(':16:error' in str(obj.exception))


def test_single_neurite_no_soma():
    string_section = u'''

                         ( (Color Yellow)
                         (Axon)
                         (Set "axons")

    ;; An commented line and some empty lines

                         (  1.2  2.7   1.0     13)  ;; Some comment
                         (  1.2  3.7   2.0     13)

                         Generated
                         )  ;  End of tree'''
    n = nm.load_neuron(string_section, reader='asc')

    assert_array_equal(n.soma.points, np.empty((0, 4)))
    nt.assert_equal(len(n.neurites), 1)
    assert_array_equal(n.neurites[0].points,
                       np.array([[1.2, 2.7, 1.0, 6.5],
                                 [1.2, 3.7, 2.0, 6.5]], dtype=np.float32))


def test_skip_header():
    '''Test that the header does not cause any issue'''
    str_neuron = '''(FilledCircle
                         (Color RGB (64, 0, 128))
                         (Name "Marker 11")
                         (Set "axons")
                         ( -189.59    55.67    28.68     0.12)  ; 1
                         )  ;  End of markers

                         ((Color Yellow)
                         (Axon)
                         (Set "axons")
                         (  1.2  2.7   1.0     13)
                         (  1.2  3.7   2.0     13)
                         )'''

    n = nm.load_neuron(str_neuron, reader='asc')
    nt.assert_equal(len(n.neurites), 1)
    assert_array_equal(n.neurites[0].points,
                       np.array([[1.2, 2.7, 1.0, 6.5],
                                 [1.2, 3.7, 2.0, 6.5]], dtype=np.float32))


without_duplicate = '''
                     ((Dendrite)
                      (3 -4 0 2)
                      (3 -6 0 2)
                      (3 -8 0 2)
                      (3 -10 0 2)
                      (
                        (0 -10 0 2)
                        (-3 -10 0 2)
                        |
                        (6 -10 0 2)
                        (9 -10 0 2)
                      )
                      )
                     '''

with_duplicate = u'''
                     ((Dendrite)
                      (3 -4 0 2)
                      (3 -6 0 2)
                      (3 -8 0 2)
                      (3 -10 0 2)
                      (
                        (3 -10 0 2) ; duplicate
                        (0 -10 0 2)
                        (-3 -10 0 2)
                        |
                        (3 -10 0 2) ; duplicate
                        (6 -10 0 2)
                        (9 -10 0 2)
                      )
                      )
                     '''


def test_read_with_duplicates():
    '''Section points are duplicated in the file'''
# what I think the
# https://developer.humanbrainproject.eu/docs/projects/morphology-documentation/0.0.2/h5v1.html
# would look like
    n = load_neuron(StringIO(with_duplicate), reader='asc')

    nt.assert_equal(len(n.neurites), 1)

    assert_array_equal(n.neurites[0].points,
                       # Duplicate points are not present
                       [[3, -4, 0, 1],
                        [3, -6, 0, 1],
                        [3, -8, 0, 1],
                        [3, -10, 0, 1],
                        [0, -10, 0, 1],
                        [-3, -10, 0, 1],
                        [6, -10, 0, 1],
                        [9, -10, 0, 1]])

    assert_array_equal(n.neurites[0].root_node.points,
                       [[3, -4, 0, 1],
                        [3, -6, 0, 1],
                        [3, -8, 0, 1],
                        [3, -10, 0, 1]])

    assert_array_equal(n.neurites[0].root_node.children[0].points,
                       [[3, -10, 0, 1],
                        [0, -10, 0, 1],
                        [-3, -10, 0, 1]])

    assert_array_equal(n.neurites[0].root_node.children[1].points,
                       [[3, -10, 0, 1],
                        [6, -10, 0, 1],
                        [9, -10, 0, 1]])


def test_read_without_duplicates():
    n_with_duplicate = load_neuron(with_duplicate, reader='asc')
    n_without_duplicate = load_neuron(without_duplicate, reader='asc')

    assert_array_equal(n_with_duplicate.neurites[0].root_node.children[0].points,
                       n_without_duplicate.neurites[0].root_node.children[0].points)

    assert_array_equal(n_with_duplicate.neurites[0].points,
                       n_without_duplicate.neurites[0].points)


# def test_broken_duplicate():
#     with assert_raises(RawDataError) as obj:
#         load_neuron(('asc',
#                      '''
#                      ((Dendrite)
#                       (3 -4 0 2)
#                       (3 -6 0 2)
#                       (3 -8 0 2)
#                       (3 -10 0 2)
#                       (
#                         (3 -10 0 40) ; <-- duplicate with different radii
#                         (0 -10 0 2)
#                         (-3 -10 0 2)
#                         |
#                         (3 -10 0 2) ; <-- good duplicate
#                         (6 -10 0 2)
#                         (9 -10 0 2)
#                       )
#                       )
#                      '''))

#     ok_("Parent point is duplicated but have a different radius" in str(obj.exception))


def test_unfinished_file():
    with assert_raises(RawDataError) as obj:
        load_neuron('''
                     ((Dendrite)
                      (3 -4 0 2)
                      (3 -6 0 2)
                      (3 -8 0 2)
                      (3 -10 0 2)
                      (
                        (3 -10 0 2)
                        (0 -10 0 2)
                        (-3 -10 0 2)
                        |
                     ''', reader='asc')

        ok_("Hit end of of file while consuming a neurite " in str(obj.exception))


def test_empty_sibling():
    n = load_neuron('''
                     ((Dendrite)
                      (3 -4 0 2)
                      (3 -6 0 2)
                      (3 -8 0 2)
                      (3 -10 0 2)
                      (
                        (3 -10 0 2)
                        (0 -10 0 2)
                        (-3 -10 0 2)
                        |
                       )
                      )
                 ''', reader='asc')

    assert_array_equal(n.neurites[0].points,
                       np.array([[3, -4, 0, 1],
                                 [3, -6, 0, 1],
                                 [3, -8, 0, 1],
                                 [3, -10, 0, 1],
                                 [0, -10, 0, 1],
                                 [-3, -10, 0, 1]],
                                dtype=np.float32))


# def test_single_children():
#     n = load_neuron(('asc',
#                      '''
#                      ((Dendrite)
#                       (3 -4 0 2)
#                       (3 -6 0 2)
#                       (3 -8 0 2)
#                       (3 -10 0 2)
#                       (
#                         (3 -10 0 2)
#                         (0 -10 0 2)
#                         (-3 -10 0 2)
#                        )
#                       )
#                  '''))

#     assert_array_equal(n.neurites[0].points,
#                        np.array([[3, -4, 0, 1],
#                                  [3, -6, 0, 1],
#                                  [3, -8, 0, 1],
#                                  [3, -10, 0, 1],
#                                  [0, -10, 0, 1],
#                                  [-3, -10, 0, 1]],
#                                 dtype=np.float32))

#     nt.assert_equal(len(n.sections), 2)
#     assert_array_equal(n.neurites[0].points,
#                        np.array([[3, -4, 0, 1],
#                                  [3, -6, 0, 1],
#                                  [3, -8, 0, 1],
#                                  [3, -10, 0, 1],
#                                  [0, -10, 0, 1],
#                                  [-3, -10, 0, 1]],
#                                 dtype=np.float32))


def test_markers():
    '''Test that markers do not prevent file from being read correctly'''
    n = load_neuron('''
( (Color White)  ; [10,1]
  (Dendrite)
  ( -290.87  -113.09   -16.32     2.06)  ; Root
  ( -290.87  -113.09   -16.32     2.06)  ; R, 1
  (
    ( -277.14  -119.13   -18.02     0.69)  ; R-1, 1
    ( -275.54  -119.99   -16.67     0.69)  ; R-1, 2
    (Cross  ;  [3,3]
      (Color Orange)
      (Name "Marker 3")
      ( -271.87  -121.14   -16.27     0.69)  ; 1
      ( -269.34  -122.29   -15.48     0.69)  ; 2
    )  ;  End of markers
     Normal
  |
    ( -277.80  -120.28   -19.48     0.92)  ; R-2, 1
    ( -276.65  -121.14   -20.20     0.92)  ; R-2, 2
    (Cross  ;  [3,3]
      (Color Orange)
      (Name "Marker 3")
      ( -279.41  -119.99   -18.00     0.46)  ; 1
      ( -272.98  -126.60   -21.22     0.92)  ; 2
    )  ;  End of markers
    (
      ( -267.94  -128.61   -22.57     0.69)  ; R-2-1, 1
      ( -204.90  -157.63   -42.45     0.69)  ; R-2-1, 34
      (Cross  ;  [3,3]
        (Color Orange)
        (Name "Marker 3")
        ( -223.67  -157.92   -42.45     0.69)  ; 1
        ( -222.76  -154.18   -39.90     0.69)  ; 2
      )  ;  End of markers
       Incomplete
    |
      ( -269.77  -129.47   -22.57     0.92)  ; R-2-2, 1
      ( -268.17  -130.62   -24.75     0.92)  ; R-2-2, 2
      ( -266.79  -131.77   -26.13     0.92)  ; R-2-2, 3
       Incomplete
    )  ;  End of split
  )  ;  End of split
)
''', reader='asc')

    nt.assert_equal(len(n.neurites), 1)

    res = np.array([[-290.87,  -113.09,   -16.32,     1.03],
                    [-290.87,  -113.09,   -16.32,     1.03],
                    [-277.14,  -119.13,   -18.02,     0.345],
                    [-275.54,  -119.99,   -16.67,     0.345],
                    [-277.80,  -120.28,   -19.48,     0.46],
                    [-276.65,  -121.14,   -20.20,     0.46],
                    [-267.94,  -128.61,   -22.57,     0.345],
                    [-204.90,  -157.63,   -42.45,     0.345],
                    [-269.77,  -129.47,   -22.57,     0.46],
                    [-268.17,  -130.62,   -24.75,     0.46],
                    [-266.79,  -131.77,   -26.13,     0.46]],
                   dtype=np.float32)

    assert_array_equal(n.neurites[0].points,
                       res)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Neuron classes and functions."""

from collections import deque
from itertools import chain

import numpy as np

import morphio

from neurom import morphmath
from neurom.core._soma import Soma, make_soma
from neurom.core.dataformat import COLS
from neurom.core.types import NeuriteIter, NeuriteType
from neurom.utils import memoize


class Section(object):
    '''Simple recursive tree class'''

    def __init__(self, morphio_section):
        '''The section ctor'''
        self.morphio_section = morphio_section

    @property
    def id(self):
        '''Returns the section ID'''
        return self.morphio_section.id

    @property
    def parent(self):
        '''Returns the parent section if non root section else None'''
        return None if self.morphio_section.is_root else Section(self.morphio_section.parent)

    @property
    def children(self):
        '''Returns a list of child section'''
        return [Section(child) for child in self.morphio_section.children]

    def append_section(self, section):
        '''
        Args:
            section (morphio.Section|morphio.mut.Section|Section|morphio.PointLevel): a section
        '''

        if isinstance(section, Section):
            return self.morphio_section.append_section(section.morphio_section)
        return self.morphio_section.append_section(section)

    def is_forking_point(self):
        '''Is tree a forking point?'''
        return len(self.children) > 1

    def is_bifurcation_point(self):
        '''Is tree a bifurcation point?'''
        return len(self.children) == 2

    def is_leaf(self):
        '''Is tree a leaf?'''
        return len(self.children) == 0

    def is_root(self):
        '''Is tree the root node?'''
        return self.parent is None

    def ipreorder(self):
        '''Depth-first pre-order iteration of tree nodes'''
        children = deque((self, ))
        while children:
            cur_node = children.pop()
            children.extend(reversed(cur_node.children))
            yield cur_node

    def ipostorder(self):
        '''Depth-first post-order iteration of tree nodes'''
        children = [self, ]
        seen = set()
        while children:
            cur_node = children[-1]
            if cur_node not in seen:
                seen.add(cur_node)
                children.extend(reversed(cur_node.children))
            else:
                children.pop()
                yield cur_node

    def iupstream(self):
        '''Iterate from a tree node to the root nodes'''
        t = self
        while t is not None:
            yield t
            t = t.parent

    def ileaf(self):
        '''Iterator to all leaves of a tree'''
        return filter(Tree.is_leaf, self.ipreorder())

    def iforking_point(self, iter_mode=ipreorder):
        '''Iterator to forking points. Returns a tree object.

        Parameters:
            tree: the tree over which to iterate
            iter_mode: iteration mode. Default: ipreorder.
        '''
        return filter(Tree.is_forking_point, iter_mode(self))

    def ibifurcation_point(self, iter_mode=ipreorder):
        '''Iterator to bifurcation points. Returns a tree object.

        Parameters:
            tree: the tree over which to iterate
            iter_mode: iteration mode. Default: ipreorder.
        '''
        return filter(Tree.is_bifurcation_point, iter_mode(self))

    def __eq__(self, other):
        return self.morphio_section == other.morphio_section

    def __hash__(self):
        return hash(self.id)

    def __nonzero__(self):
        return bool(self.children)

    __bool__ = __nonzero__

    @property
    def points(self):
        '''Returns the section list of points the NeuroM way (points + radius)'''
        return np.concatenate((self.morphio_section.points,
                               self.morphio_section.diameters[:, np.newaxis] / 2.),
                              axis=1)

    @points.setter
    def points(self, value):
        '''Set the points'''
        self.morphio_section.points = np.copy(value[:, COLS.XYZ])
        self.morphio_section.diameters = np.copy(value[:, COLS.R]) * 2

    @property
    def type(self):
        '''Returns the section type'''
        return self.morphio_section.type

    # TODO: Should we have a @type.setter ?

    @property
    def length(self):
        '''Return the path length of this section.'''
        return morphmath.section_length(self.points)

    @property
    def area(self):
        '''Return the surface area of this section.

        The area is calculated from the segments, as defined by this
        section's points
        '''
        return sum(morphmath.segment_area(s) for s in iter_segments(self))

    @property
    def volume(self):
        '''Return the volume of this section.

        The volume is calculated from the segments, as defined by this
        section's points
        '''
        return sum(morphmath.segment_volume(s) for s in iter_segments(self))

    def __str__(self):
        return 'Section(id=%s, type=%s, n_points=%s) <parent: %s, nchildren: %d>' % \
            (self.id, self.type, len(self.points), self.parent, len(self.children))

    __repr__ = __str__


Tree = Section

# NRN simulator iteration order
# See:
# https://github.com/neuronsimulator/nrn/blob/2dbf2ebf95f1f8e5a9f0565272c18b1c87b2e54c/share/lib/hoc/import3d/import3d_gui.hoc#L874
NRN_ORDER = {NeuriteType.soma: 0,
             NeuriteType.axon: 1,
             NeuriteType.basal_dendrite: 2,
             NeuriteType.apical_dendrite: 3,
             NeuriteType.undefined: 4}


def iter_neurites(obj, mapfun=None, filt=None, neurite_order=NeuriteIter.FileOrder):
    """Iterator to a neurite, neuron or neuron population.

    Applies optional neurite filter and mapping functions.

    Arguments:
        obj: a neurite, neuron or neuron population.
        mapfun: optional neurite mapping function.
        filt: optional neurite filter function.
        neurite_order (NeuriteIter): order upon which neurites should be iterated
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical

    Examples:
        Get the number of points in each neurite in a neuron population

        >>> from neurom.core import iter_neurites
        >>> n_points = [n for n in iter_neurites(pop, lambda x : len(x.points))]

        Get the number of points in each axon in a neuron population

        >>> import neurom as nm
        >>> from neurom.core import iter_neurites
        >>> filter = lambda n : n.type == nm.AXON
        >>> mapping = lambda n : len(n.points)
        >>> n_points = [n for n in iter_neurites(pop, mapping, filter)]
    """
    neurites = ((obj,) if isinstance(obj, Neurite) else
                obj.neurites if hasattr(obj, 'neurites') else obj)
    if neurite_order == NeuriteIter.NRN:
        last_position = max(NRN_ORDER.values()) + 1
        neurites = sorted(neurites, key=lambda neurite: NRN_ORDER.get(neurite.type, last_position))

    neurite_iter = iter(neurites) if filt is None else filter(filt, neurites)
    return neurite_iter if mapfun is None else map(mapfun, neurite_iter)


def iter_sections(neurites,
                  iterator_type=Tree.ipreorder,
                  neurite_filter=None,
                  neurite_order=NeuriteIter.FileOrder):
    """Iterator to the sections in a neurite, neuron or neuron population.

    Arguments:
        neurites: neuron, population, neurite, or iterable containing neurite objects
        iterator_type: section iteration order within a given neurite. Must be one of:
            Tree.ipreorder: Depth-first pre-order iteration of tree nodes
            Tree.ipostorder: Depth-first post-order iteration of tree nodes
            Tree.iupstream: Iterate from a tree node to the root nodes
            Tree.ibifurcation_point: Iterator to bifurcation points
            Tree.ileaf: Iterator to all leaves of a tree

        neurite_filter: optional top level filter on properties of neurite neurite objects.
        neurite_order (NeuriteIter): order upon which neurites should be iterated
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical


    Examples:
        Get the number of points in each section of all the axons in a neuron population

        >>> import neurom as nm
        >>> from neurom.core import ites_sections
        >>> filter = lambda n : n.type == nm.AXON
        >>> n_points = [len(s.points) for s in iter_sections(pop,  neurite_filter=filter)]
    """
    return chain.from_iterable(
        iterator_type(neurite.root_node) for neurite in
        iter_neurites(neurites, filt=neurite_filter, neurite_order=neurite_order))


def iter_segments(obj, neurite_filter=None, neurite_order=NeuriteIter.FileOrder):
    """Return an iterator to the segments in a collection of neurites.

    Arguments:
        obj: neuron, population, neurite, section, or iterable containing neurite objects
        neurite_filter: optional top level filter on properties of neurite neurite objects
        neurite_order: order upon which neurite should be iterated. Values:
            - NeuriteIter.FileOrder: order of appearance in the file
            - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical

    Note:
        This is a convenience function provided for generic access to
        neuron segments. It may have a performance overhead WRT custom-made
        segment analysis functions that leverage numpy and section-wise iteration.
    """
    sections = iter((obj,) if isinstance(obj, Section) else
                    iter_sections(obj,
                                  neurite_filter=neurite_filter,
                                  neurite_order=neurite_order))

    return chain.from_iterable(zip(sec.points[:-1], sec.points[1:])
                               for sec in sections)


def graft_neuron(section):
    '''Returns a neuron starting at section'''
    assert isinstance(section, Section)
    m = morphio.mut.Morphology()
    m.append_root_section(section.morphio_section)
    return Neuron(m)


class Neurite(object):
    '''Class representing a neurite tree'''

    def __init__(self, root_node):
        self.morphio_root_node = root_node

    @property
    def root_node(self):
        '''Returns the first section of the neurite'''
        return Section(self.morphio_root_node)

    @property
    def type(self):
        '''Returns the type of the root node'''
        return self.root_node.type

    @property
    # @memoize
    def points(self):
        '''Returns unordered array with all the points in this neurite

        Note: Duplicate points at section bifurcations are removed'''

        # Neurite first point must be added manually
        _ptr = list(chain([Section(self.morphio_root_node).points[0][COLS.XYZR]],
                          [v for s in self.morphio_root_node.iter()
                           for v in Section(s).points[1:, COLS.XYZR]]))
        return np.array(_ptr)

    @property
    @memoize
    def length(self):
        """Returns the total length of this neurite.

        The length is defined as the sum of lengths of the sections.
        """
        return sum(s.length for s in self.iter_sections())

    @property
    @memoize
    def area(self):
        """Returns the surface area of this neurite.

        The area is defined as the sum of area of the sections.
        """
        return sum(s.area for s in self.iter_sections())

    @property
    @memoize
    def volume(self):
        """Returns the volume of this neurite.

        The volume is defined as the sum of volumes of the sections.
        """
        return sum(s.volume for s in self.iter_sections())

    def iter_sections(self, order=Tree.ipreorder, neurite_order=NeuriteIter.FileOrder):
        """Iteration over section nodes.

        Arguments:
            order: section iteration order within a given neurite. Must be one of:
                Tree.ipreorder: Depth-first pre-order iteration of tree nodes
                Tree.ipreorder: Depth-first post-order iteration of tree nodes
                Tree.iupstream: Iterate from a tree node to the root nodes
                Tree.ibifurcation_point: Iterator to bifurcation points
                Tree.ileaf: Iterator to all leaves of a tree

            neurite_order: order upon which neurites should be iterated. Values:
                - NeuriteIter.FileOrder: order of appearance in the file
                - NeuriteIter.NRN: NRN simulator order: soma -> axon -> basal -> apical
        """
        return iter_sections(self, iterator_type=order, neurite_order=neurite_order)

    # def __deepcopy__(self, memo):
    #     """Deep copy of neurite object"""
    #     return Neurite(deepcopy(self.morphio_root_node, memo))

    def __nonzero__(self):
        return bool(self.morphio_root_node)

    def __eq__(self, other):
        return self.type == other.type and self.morphio_root_node.id == other.morphio_root_node.id

    def __hash__(self):
        return hash((self.type, self.morphio_root_node))

    __bool__ = __nonzero__

    def __str__(self):
        """Return a string representation."""
        return 'Neurite <type: %s>' % self.type

    __repr__ = __str__


class Neuron(morphio.mut.Morphology):
    """Class representing a simple neuron"""

    def __init__(self, filename, name=None):
        """Neuron ctor

        Args:
            filename (str|Path): a filename
            name (str): a option neuron name
        """
        super(Neuron, self).__init__(filename)
        self.name = name if name else 'Neuron'
        self.morphio_soma = super(Neuron, self).soma

        self.neurom_soma = make_soma(self.morphio_soma)

    def __copy__(self):
        return Neuron(self)

    @property
    def soma(self):
        """Return the soma"""
        return self.neurom_soma

    @soma.setter
    def soma(self, value):
        """Set the soma"""
        self.neurom_soma = value

    def __str__(self):
        """Return a string representation."""
        return 'Neuron <soma: %s, n_neurites: %d>' % \
            ('1', len(self.neurites))

    @property
    def neurites(self):
        """Returns the list of neurites"""
        return [Neurite(root_section) for root_section in self.root_sections]

    @property
    def sections(self):
        """Returns an array of all sections (excluding the soma)"""
        return list(iter_sections(self))

    @property
    def points(self):
        """Returns the list of points"""
        return np.concatenate(
            [section.points for section in iter_sections(self)])

    def transform(self, trans):
        """Return a copy of this neuron with a 3D transformation applied"""
        obj = Neuron(self)
        obj.morphio_soma.points = trans(obj.morphio_soma.points)

        for section in obj.sections:
            section.morphio_section.points = trans(section.morphio_section.points)
        return obj

    __repr__ = __str__



#!/usr/bin/env python
# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Load and view multiple somas."""

from pathlib import Path

from neurom import load_neuron
import neurom.view.common as common
import matplotlib.pyplot as plt
import numpy as np

DATA_PATH = Path(__file__).parent.parent / 'test_data'
SWC_PATH = Path(DATA_PATH, 'swc')


def random_color():
    """Random color generation."""
    return np.random.rand(3, 1)


def plot_somas(somas):
    """Plot set of somas on same figure as spheres, each with different color."""
    _, ax = common.get_figure(new_fig=True, subplot=111,
                              params={'projection': '3d', 'aspect': 'equal'})
    for s in somas:
        common.plot_sphere(ax, s.center, s.radius, color=random_color(), alpha=1)
    plt.show()


if __name__ == '__main__':
    #  define set of files containing relevant neurons
    file_nms = [Path(SWC_PATH, file_nm) for file_nm in ['Soma_origin.swc',
                                                                'Soma_translated_1.swc',
                                                                'Soma_translated_2.swc']]

    # load from file and plot
    sms = [load_neuron(file_nm).soma for file_nm in file_nms]
    plot_somas(sms)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test neurom._neuronfunc functionality."""
from pathlib import Path
import tempfile
import warnings
from io import StringIO

import numpy as np
from nose import tools as nt
from numpy.testing import (assert_almost_equal, assert_array_almost_equal,
                           assert_array_equal)

from neurom import NeuriteType, load_neuron
from neurom.core.population import Population

# NOTE: The 'bf' alias is used in the fst/tests modules
# Do NOT change it.
# TODO: If other neurom.features are imported,
# the should use the aliasing used in fst/tests module files
from neurom.features import neuronfunc as _nf

DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data'
H5_PATH = DATA_PATH / 'h5/v1'
NRN = load_neuron(Path(H5_PATH, 'Neuron.h5'))

SWC_PATH = DATA_PATH / 'swc'
SIMPLE = load_neuron(Path(SWC_PATH, 'simple.swc'))
SIMPLE_TRUNK = load_neuron(Path(SWC_PATH, 'simple_trunk.swc'))
SWC_NRN = load_neuron(Path(SWC_PATH, 'Neuron.swc'))
with warnings.catch_warnings(record=True):
    SWC_NRN_3PT = load_neuron(Path(SWC_PATH, 'soma', 'three_pt_soma.swc'))


def test_soma_volume():
    with warnings.catch_warnings(record=True):
        # SomaSinglePoint
        ret = _nf.soma_volume(SIMPLE)
        assert_almost_equal(ret, 4.1887902047863905)
        # SomaCylinders
        ret = _nf.soma_volume(SWC_NRN)
        assert_almost_equal(ret, 0.010726068245337955)
        # SomaSimpleContour
        ret = _nf.soma_volume(NRN)
        assert_almost_equal(ret, 0.0033147000251481135)
        # SomaNeuromorphoThreePointCylinders
        ret = _nf.soma_volume(SWC_NRN_3PT)
        assert_almost_equal(ret, 50.26548245743669)


def test_soma_volumes():
    with warnings.catch_warnings(record=True):
        ret = _nf.soma_volumes(SIMPLE)
        nt.eq_(ret, [4.1887902047863905, ])


def test_soma_surface_area():
    ret = _nf.soma_surface_area(SIMPLE)
    nt.eq_(ret, 12.566370614359172)


def test_soma_surface_areas():
    ret = _nf.soma_surface_areas(SIMPLE)
    nt.eq_(ret, [12.566370614359172, ])


def test_soma_radii():
    ret = _nf.soma_radii(SIMPLE)
    nt.eq_(ret, [1., ])


def test_trunk_section_lengths():
    ret = _nf.trunk_section_lengths(SIMPLE)
    nt.eq_(ret, [5.0, 4.0])


def test_trunk_origin_radii():
    ret = _nf.trunk_origin_radii(SIMPLE)
    nt.eq_(ret, [1.0, 1.0])


def test_trunk_origin_azimuths():
    ret = _nf.trunk_origin_azimuths(SIMPLE)
    nt.eq_(ret, [0.0, 0.0])


def test_trunk_angles():
    ret = _nf.trunk_angles(SIMPLE_TRUNK)
    assert_array_almost_equal(ret, [np.pi/2, np.pi/2, np.pi/2, np.pi/2])
    ret = _nf.trunk_angles(SIMPLE_TRUNK, neurite_type=NeuriteType.basal_dendrite)
    assert_array_almost_equal(ret, [np.pi, np.pi])
    ret = _nf.trunk_angles(SIMPLE_TRUNK, neurite_type=NeuriteType.axon)
    assert_array_almost_equal(ret, [0.0])
    ret = _nf.trunk_angles(SIMPLE, neurite_type=NeuriteType.apical_dendrite)
    assert_array_almost_equal(ret, [])


def test_trunk_vectors():
    ret = _nf.trunk_vectors(SIMPLE_TRUNK)
    assert_array_equal(ret[0], [0., -1.,  0.])
    assert_array_equal(ret[1], [1.,  0.,  0.])
    assert_array_equal(ret[2], [-1.,  0.,  0.])
    assert_array_equal(ret[3], [0.,  1.,  0.])
    ret = _nf.trunk_vectors(SIMPLE_TRUNK, neurite_type=NeuriteType.axon)
    assert_array_equal(ret[0], [0., -1.,  0.])


def test_trunk_origin_elevations():
    n0 = load_neuron(StringIO(u"""
    1 1 0 0 0 4 -1
    2 3 1 0 0 2 1
    3 3 2 1 1 2 2
    4 3 0 1 0 2 1
    5 3 1 2 1 2 4
    """), reader='swc')

    n1 = load_neuron(StringIO(u"""
    1 1 0 0 0 4 -1
    2 3 0 -1 0 2 1
    3 3 -1 -2 -1 2 2
    """), reader='swc')

    pop = Population([n0, n1])
    assert_array_almost_equal(_nf.trunk_origin_elevations(pop),
                       [0.0, np.pi/2., -np.pi/2.])

    assert_array_almost_equal(_nf.trunk_origin_elevations(pop, neurite_type=NeuriteType.basal_dendrite),
                       [0.0, np.pi/2., -np.pi/2.])

    assert_array_almost_equal(_nf.trunk_origin_elevations(pop, neurite_type=NeuriteType.axon),
                       [])

    assert_array_almost_equal(_nf.trunk_origin_elevations(pop, neurite_type=NeuriteType.apical_dendrite),
[])


@nt.raises(Exception)
def test_trunk_elevation_zero_norm_vector_raises():
    _nf.trunk_origin_elevations(SWC_NRN)


def test_sholl_crossings_simple():
    center = SIMPLE.soma.center
    radii = []
    nt.eq_(list(_nf.sholl_crossings(SIMPLE, center, radii=radii)),
           [])

    radii = [1.0]
    nt.eq_([2],
           list(_nf.sholl_crossings(SIMPLE, center, radii=radii)))

    radii = [1.0, 5.1]
    nt.eq_([2, 4],
           list(_nf.sholl_crossings(SIMPLE, center, radii=radii)))

    radii = [1., 4., 5.]
    nt.eq_([2, 4, 5],
           list(_nf.sholl_crossings(SIMPLE, center, radii=radii)))


def load_swc(string):
    with tempfile.NamedTemporaryFile(prefix='test_neuron_func', mode='w', suffix='.swc') as fd:
        fd.write(string)
        fd.flush()
        return load_neuron(fd.name)


def test_sholl_analysis_custom():
    # recreate morphs from Fig 2 of
    # http://dx.doi.org/10.1016/j.jneumeth.2014.01.016
    radii = np.arange(10, 81, 10)
    center = 0, 0, 0
    morph_A = load_swc("""\
 1 1   0  0  0 1. -1
 2 3   0  0  0 1.  1
 3 3  80  0  0 1.  2
 4 4   0  0  0 1.  1
 5 4 -80  0  0 1.  4""")
    nt.eq_(list(_nf.sholl_crossings(morph_A, center, radii=radii)),
           [2, 2, 2, 2, 2, 2, 2, 2])

    morph_B = load_swc("""\
 1 1   0   0  0 1. -1
 2 3   0   0  0 1.  1
 3 3  35   0  0 1.  2
 4 3  51  10  0 1.  3
 5 3  51   5  0 1.  3
 6 3  51   0  0 1.  3
 7 3  51  -5  0 1.  3
 8 3  51 -10  0 1.  3
 9 4 -35   0  0 1.  2
10 4 -51  10  0 1.  9
11 4 -51   5  0 1.  9
12 4 -51   0  0 1.  9
13 4 -51  -5  0 1.  9
14 4 -51 -10  0 1.  9
                       """)
    nt.eq_(list(_nf.sholl_crossings(morph_B, center, radii=radii)),
           [2, 2, 2, 10, 10, 0, 0, 0])

    morph_C = load_swc("""\
 1 1   0   0  0 1. -1
 2 3   0   0  0 1.  1
 3 3  65   0  0 1.  2
 4 3  85  10  0 1.  3
 5 3  85   5  0 1.  3
 6 3  85   0  0 1.  3
 7 3  85  -5  0 1.  3
 8 3  85 -10  0 1.  3
 9 4  65   0  0 1.  2
10 4  85  10  0 1.  9
11 4  85   5  0 1.  9
12 4  85   0  0 1.  9
13 4  85  -5  0 1.  9
14 4  85 -10  0 1.  9
                       """)
    nt.eq_(list(_nf.sholl_crossings(morph_C, center, radii=radii)),
           [2, 2, 2, 2, 2, 2, 10, 10])
    # view.neuron(morph_C)[0].savefig('foo.png')



"""Define the public 'draw' function to be used to draw morphology using plotly."""
from __future__ import absolute_import  # prevents name clash with local plotly module
from itertools import chain

import numpy as np

try:
    import plotly.graph_objs as go
    from plotly.offline import plot, iplot, init_notebook_mode
except ImportError:
    raise ImportError(
        'neurom[plotly] is not installed. Please install it by doing: pip install neurom[plotly]')

from neurom import COLS, iter_segments, iter_neurites
from neurom.core import Neuron
from neurom.view.view import TREE_COLOR


def draw(obj, plane='3d', inline=False, **kwargs):
    """Draw the object using the given plane.

    obj (neurom.Neuron, neurom.Tree): neuron or tree
    plane (str): a string representing the 2D plane (example: 'xy')
                 or '3d', '3D' for a 3D view

    inline (bool): must be set to True for interactive ipython notebook plotting
    """
    if plane.lower() == '3d':
        return _plot_neuron3d(obj, inline, **kwargs)
    return _plot_neuron(obj, plane, inline, **kwargs)


def _plot_neuron(neuron, plane, inline, **kwargs):
    return _plotly(neuron, plane=plane, title='neuron-2D', inline=inline, **kwargs)


def _plot_neuron3d(neuron, inline, **kwargs):
    """Generates a figure of the neuron, that contains a soma and a list of trees."""
    return _plotly(neuron, plane='3d', title='neuron-3D', inline=inline, **kwargs)


def _make_trace(neuron, plane):
    """Create the trace to be plotted."""
    for neurite in iter_neurites(neuron):
        segments = list(iter_segments(neurite))

        segs = [(s[0][COLS.XYZ], s[1][COLS.XYZ]) for s in segments]

        coords = dict(x=list(chain.from_iterable((p1[0], p2[0], None) for p1, p2 in segs)),
                      y=list(chain.from_iterable((p1[1], p2[1], None) for p1, p2 in segs)),
                      z=list(chain.from_iterable((p1[2], p2[2], None) for p1, p2 in segs)))

        color = TREE_COLOR.get(neurite.root_node.type, 'black')
        if plane.lower() == '3d':
            plot_fun = go.Scatter3d
        else:
            plot_fun = go.Scatter
            coords = dict(x=coords[plane[0]], y=coords[plane[1]])
        yield plot_fun(
            line=dict(color=color, width=2),
            mode='lines',
            **coords
        )


def _fill_soma_data(neuron, data, plane):
    """Fill soma data if 3D plot and returns soma_2d in all cases."""
    if not isinstance(neuron, Neuron):
        return []

    if plane != '3d':
        soma_2d = [
            # filled circle
            {
                'type': 'circle',
                'xref': 'x',
                'yref': 'y',
                'fillcolor': 'rgba(50, 171, 96, 0.7)',
                'x0': neuron.soma.center[0] - neuron.soma.radius,
                'y0': neuron.soma.center[1] - neuron.soma.radius,
                'x1': neuron.soma.center[0] + neuron.soma.radius,
                'y1': neuron.soma.center[1] + neuron.soma.radius,

                'line': {
                    'color': 'rgba(50, 171, 96, 1)',
                },
            },
        ]

    else:
        soma_2d = []
        point_count = 100  # Enough points so that the surface looks like a sphere
        theta = np.linspace(0, 2 * np.pi, point_count)
        phi = np.linspace(0, np.pi, point_count)
        r = neuron.soma.radius
        data.append(
            go.Surface(
                x=r * np.outer(np.cos(theta), np.sin(phi)) + neuron.soma.center[0],
                y=r * np.outer(np.sin(theta), np.sin(phi)) + neuron.soma.center[1],
                z=r * np.outer(np.ones(point_count), np.cos(phi)) + neuron.soma.center[2],
                cauto=False,
                surfacecolor=['black'] * len(phi),
                showscale=False,
            )
        )
    return soma_2d


def get_figure(neuron, plane, title):
    """Returns the plotly figure containing the neuron."""
    data = list(_make_trace(neuron, plane))
    axis = dict(
        gridcolor='rgb(255, 255, 255)',
        zerolinecolor='rgb(255, 255, 255)',
        showbackground=True,
        backgroundcolor='rgb(230, 230,230)'
    )

    soma_2d = _fill_soma_data(neuron, data, plane)

    layout = dict(
        autosize=True,
        title=title,
        scene=dict(  # This is used for 3D plots
            xaxis=axis, yaxis=axis, zaxis=axis,
            camera=dict(up=dict(x=0, y=0, z=1), eye=dict(x=-1.7428, y=1.0707, z=0.7100,)),
            aspectmode='data'
        ),
        yaxis=dict(scaleanchor="x"),  # This is used for 2D plots
        shapes=soma_2d,
    )

    res = dict(data=data, layout=layout)
    return res


def _plotly(neuron, plane, title, inline, **kwargs):
    fig = get_figure(neuron, plane, title)

    plot_fun = iplot if inline else plot
    if inline:
        init_notebook_mode(connected=True)  # pragma: no cover
    plot_fun(fig, filename=title + '.html', **kwargs)
    return fig



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Type enumerations."""

from enum import unique


from neurom import NeuriteType
from neurom.utils import OrderedEnum

NeuriteType.name = property(lambda self: str(self).split('.')[-1])


@unique
class NeuriteIter(OrderedEnum):
    """Neurite iteration orders."""
    FileOrder = 1  # Order in which neurites appear in the file

    # NRN simulator order: soma -> axon -> basal -> apical
    # Coming from:
    # https://github.com/neuronsimulator/nrn/blob/2dbf2ebf95f1f8e5a9f0565272c18b1c87b2e54c/share/lib/hoc/import3d/import3d_gui.hoc#L874
    NRN = 2


NEURITES = (NeuriteType.all,
            NeuriteType.axon,
            NeuriteType.apical_dendrite,
            NeuriteType.basal_dendrite)

ROOT_ID = -1


def tree_type_checker(*ref):
    """Tree type checker functor.

    Returns:
        Functor that takes a tree, and returns true if that tree matches any of
        NeuriteTypes in ref

    Ex:
        >>> from neurom.core.types import NeuriteType, tree_type_checker
        >>> tree_filter = tree_type_checker(NeuriteType.axon, NeuriteType.basal_dendrite)
        >>> nrn.i_neurites(tree.isegment, tree_filter=tree_filter)
    """
    ref = tuple(ref)
    if NeuriteType.all in ref:
        def check_tree_type(_):
            """Always returns true."""
            return True
    else:
        def check_tree_type(tree):
            """Check whether tree has the same type as ref.

            Returns:
                True if ref in the same type as tree.type or ref is NeuriteType.all
            """
            return tree.type in ref

    return check_tree_type


def dendrite_filter(n):
    """Select only dendrites."""
    return n.type == NeuriteType.basal_dendrite or n.type == NeuriteType.apical_dendrite


def axon_filter(n):
    """Select only axons."""
    return n.type == NeuriteType.axon



"""Legacy module, replaced by neurom.features."""
from warnings import warn
from neurom.features import neuritefunc as _neuritefunc
from neurom.features import neuronfunc as _neuronfunc
from neurom.features import sectionfunc

from neurom.features import NEURITEFEATURES, NEURONFEATURES, get, register_neurite_feature
warn('neurom.fst is being deprecated and will be removed in NeuroM v1.5.0,'
     ' replace it by neurom.features', DeprecationWarning)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Dendrogram helper functions and class."""
import numpy as np

from neurom import NeuriteType
from neurom.core import Neurite, Neuron
from neurom.core.dataformat import COLS
from neurom.morphmath import interval_lengths


class Dendrogram(object):
    """Dendrogram."""

    def __init__(self, neurom_section):
        """Dendrogram for NeuroM section tree.

        Args:
            neurom_section (Neurite|Neuron|Section): tree to build dendrogram for.
        """
        if isinstance(neurom_section, Neuron):
            self.neurite_type = NeuriteType.soma
            self.height = 1
            self.width = 1
            self.coords = self.get_coords(
                np.array([0, self.height]), np.array([.5 * self.width, .5 * self.width]))
            self.children = [Dendrogram(neurite.root_node) for neurite in neurom_section.neurites]
        else:
            if isinstance(neurom_section, Neurite):
                neurom_section = neurom_section.root_node
            lengths = interval_lengths(neurom_section.points, prepend_zero=True)
            radii = neurom_section.points[:, COLS.R]

            self.neurite_type = neurom_section.type
            self.height = np.sum(lengths)
            self.width = 2 * np.max(radii)
            self.coords = Dendrogram.get_coords(lengths, radii)
            self.children = [Dendrogram(child) for child in neurom_section.children]

    @staticmethod
    def get_coords(segment_lengths, segment_radii):
        """Coordinates of dendrogram as polygon with respect to (0, 0) origin.

        Args:
            segment_lengths: lengths of dendrogram segments
            segment_radii: radii of dendrogram segments

        Returns:
            (N,2) array of 2D x,y coordinates of Dendrogram polygon. N is the number of vertices.
        """
        y_coords = np.cumsum(segment_lengths)
        x_left_coords = -segment_radii
        x_right_coords = segment_radii
        left_coords = np.hstack((x_left_coords[:, np.newaxis], y_coords[:, np.newaxis]))
        right_coords = np.hstack((x_right_coords[:, np.newaxis], y_coords[:, np.newaxis]))
        right_coords = np.flip(right_coords, axis=0)
        return np.vstack((left_coords, right_coords))


def layout_dendrogram(dendrogram, origin):
    """Lays out dendrogram as an aesthetical pleasing tree.

    Args:
        dendrogram (Dendrogram): dendrogram
        origin (np.array): xy coordinates of layout origin

    Returns:
        Dict of positions per each dendrogram node. When placed in those positions, dendrogram nodes
        will represent a nice tree structure.
    """

    class _PositionedDendrogram(object):
        """Wrapper around dendrogram that allows to lay it out.

        The layout happens only in X coordinates. Children's Y coordinate is just a parent's Y
         + parent's height. Algorithm is that we calculate bounding rectangle width of each
         dendrogram's subtree. This width is a sum of all children widths calculated recursively
         in `total_width`. If no children then the width is the dendrogram's width. After the
         calculation we start to lay out. Each child gets its X coordinate as:
         parent's X + previous sibling children widths + half of this child's width.
        """
        HORIZONTAL_PADDING = 2

        def __init__(self, dendrogram):
            self.dendrogram = dendrogram
            self.children = [_PositionedDendrogram(child) for child in dendrogram.children]
            self.total_width = self.dendrogram.width
            if self.children:
                children_width = np.sum([child.total_width for child in self.children])
                children_width += self.HORIZONTAL_PADDING * (len(self.children) - 1)
                self.total_width = max(self.total_width, children_width)

        def position_at(self, origin):  # pylint: disable=missing-docstring
            positions = {self.dendrogram: origin}
            if self.children:
                end_point = origin + [0, self.dendrogram.height]
                left_bottom_offset = [-.5 * self.total_width, 0]
                children_origin = end_point + left_bottom_offset
                for child in self.children:
                    child_origin = children_origin + [.5 * child.total_width, 0]
                    positions.update(child.position_at(child_origin))
                    children_origin += [child.total_width + self.HORIZONTAL_PADDING, 0]
            return positions

    return _PositionedDendrogram(dendrogram).position_at(origin)


def get_size(positions):
    """Get the size of bounding rectangle that embodies positions.

    Args:
        positions (dict of Dendrogram: np.array): positions xy coordinates of dendrograms

    Returns:
        Tuple of width and height of bounding rectangle.
    """
    max_y_list = [dendrogram.height + coords[1] for dendrogram, coords in positions.items()]
    coords = np.array(list(positions.values()))
    width = np.max(coords[:, 0]) - np.min(coords[:, 0])
    height = np.max(max_y_list) - np.min(coords[:, 1])
    return width, height


def move_positions(positions, to_origin):
    """Move positions to a new origin.

    Args:
        positions (dict of Dendrogram: np.array): positions
        to_origin (np.array): where to move. np.array of (2,) shape for x,y coordindates.

    Returns:
        Moved positions.
    """
    to_origin = np.asarray(to_origin)
    return {dendrogram: position + to_origin for dendrogram, position in positions.items()}



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""NeuroM, lightweight and fast.

Examples:
    Obtain some morphometrics

    >>> ap_seg_len = features.get('segment_lengths', nrn, neurite_type=neurom.APICAL_DENDRITE)
    >>> ax_sec_len = features.get('section_lengths', nrn, neurite_type=neurom.AXON)
"""

from functools import partial, update_wrapper
import numpy as _np

from neurom.features import neuritefunc as _nrt
from neurom.features import neuronfunc as _nrn
from neurom.core import NeuriteType as _ntype
from neurom.core import iter_neurites as _ineurites
from neurom.core.types import tree_type_checker as _is_type
from neurom.exceptions import NeuroMError

_partition_asymmetry_length = partial(_nrt.partition_asymmetries, variant='length')
update_wrapper(_partition_asymmetry_length, _nrt.partition_asymmetries)

NEURITEFEATURES = {
    'total_length': _nrt.total_length,
    'total_length_per_neurite': _nrt.total_length_per_neurite,
    'neurite_lengths': _nrt.total_length_per_neurite,
    'terminal_path_lengths_per_neurite': _nrt.terminal_path_lengths_per_neurite,
    'section_lengths': _nrt.section_lengths,
    'section_term_lengths': _nrt.section_term_lengths,
    'section_bif_lengths': _nrt.section_bif_lengths,
    'neurite_volumes': _nrt.total_volume_per_neurite,
    'neurite_volume_density': _nrt.neurite_volume_density,
    'section_volumes': _nrt.section_volumes,
    'section_areas': _nrt.section_areas,
    'section_tortuosity': _nrt.section_tortuosity,
    'section_path_distances': _nrt.section_path_lengths,
    'number_of_sections': _nrt.number_of_sections,
    'number_of_sections_per_neurite': _nrt.number_of_sections_per_neurite,
    'number_of_neurites': _nrt.number_of_neurites,
    'number_of_bifurcations': _nrt.number_of_bifurcations,
    'number_of_forking_points': _nrt.number_of_forking_points,
    'number_of_terminations': _nrt.number_of_terminations,
    'section_branch_orders': _nrt.section_branch_orders,
    'section_term_branch_orders': _nrt.section_term_branch_orders,
    'section_bif_branch_orders': _nrt.section_bif_branch_orders,
    'section_radial_distances': _nrt.section_radial_distances,
    'section_bif_radial_distances': _nrt.section_bif_radial_distances,
    'section_term_radial_distances': _nrt.section_term_radial_distances,
    'section_end_distances': _nrt.section_end_distances,
    'section_strahler_orders': _nrt.section_strahler_orders,
    'local_bifurcation_angles': _nrt.local_bifurcation_angles,
    'remote_bifurcation_angles': _nrt.remote_bifurcation_angles,
    'partition': _nrt.bifurcation_partitions,
    'partition_asymmetry': _nrt.partition_asymmetries,
    'partition_pairs': _nrt.partition_pairs,
    'partition_asymmetry_length': _partition_asymmetry_length,
    'sibling_ratio': _nrt.sibling_ratios,
    'diameter_power_relation': _nrt.diameter_power_relations,
    'number_of_segments': _nrt.number_of_segments,
    'segment_lengths': _nrt.segment_lengths,
    'segment_areas': _nrt.segment_areas,
    'segment_volumes': _nrt.segment_volumes,
    'segment_radii': _nrt.segment_radii,
    'segment_midpoints': _nrt.segment_midpoints,
    'segment_taper_rates': _nrt.segment_taper_rates,
    'segment_path_lengths': _nrt.segment_path_lengths,
    'section_taper_rates': _nrt.section_taper_rates,
    'segment_radial_distances': _nrt.segment_radial_distances,
    'segment_meander_angles': _nrt.segment_meander_angles,
    'principal_direction_extents': _nrt.principal_direction_extents,
    'total_area_per_neurite': _nrt.total_area_per_neurite,
}

NEURONFEATURES = {
    'soma_radii': _nrn.soma_radii,
    'soma_surface_areas': _nrn.soma_surface_areas,
    'soma_volumes': _nrn.soma_volumes,
    'trunk_origin_radii': _nrn.trunk_origin_radii,
    'trunk_origin_azimuths': _nrn.trunk_origin_azimuths,
    'trunk_origin_elevations': _nrn.trunk_origin_elevations,
    'trunk_section_lengths': _nrn.trunk_section_lengths,
    'trunk_angles': _nrn.trunk_angles,
    'trunk_vectors': _nrn.trunk_vectors,
    'sholl_frequency': _nrn.sholl_frequency,
}


def register_neurite_feature(name, func):
    """Register a feature to be applied to neurites.

    Arguments:
        name: name of the feature, used for access via get() function.
        func: single parameter function of a neurite.
    """
    if name in NEURITEFEATURES:
        raise NeuroMError('Attempt to hide registered feature %s' % name)

    def _fun(neurites, neurite_type=_ntype.all):
        """Wrap neurite function from outer scope and map into list."""
        return list(func(n) for n in _ineurites(neurites, filt=_is_type(neurite_type)))

    NEURONFEATURES[name] = _fun


def get(feature, obj, **kwargs):
    """Obtain a feature from a set of morphology objects.

    Arguments:
        feature(string): feature to extract
        obj: a neuron, population or neurite tree
        kwargs: parameters to forward to underlying worker functions

    Returns:
        features as a 1D or 2D numpy array.
    """
    feature = (NEURITEFEATURES[feature] if feature in NEURITEFEATURES
               else NEURONFEATURES[feature])

    return _np.array(list(feature(obj, **kwargs)))


_INDENT = ' ' * 4


def _indent(string, count):
    """Indent `string` by `count` * INDENT."""
    indent = _INDENT * count
    ret = indent + string.replace('\n', '\n' + indent)
    return ret.rstrip()


def _get_doc():
    """Get a description of all the known available features."""
    def get_docstring(func):
        """Extract doctstring, if possible."""
        docstring = ':\n'
        if func.__doc__:
            docstring += _indent(func.__doc__, 2)
        return docstring

    ret = ['\nNeurite features (neurite, neuron, neuron population):']
    ret.extend(_INDENT + '- ' + feature + get_docstring(func)
               for feature, func in sorted(NEURITEFEATURES.items()))

    ret.append('\nNeuron features (neuron, neuron population):')
    ret.extend(_INDENT + '- ' + feature + get_docstring(func)
               for feature, func in sorted(NEURONFEATURES.items()))

    return '\n'.join(ret)


get.__doc__ += _indent('\nFeatures:\n', 1) + _indent(_get_doc(), 2)  # pylint: disable=no-member



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""NeuroM helper utilities."""
from enum import Enum
import json
import warnings
from functools import partial, update_wrapper, wraps

import numpy as np


class memoize(object):
    """cache the return value of a method.

    This class is meant to be used as a decorator of methods. The return value
    from a given method invocation will be cached on the instance whose method
    was invoked. All arguments passed to a method decorated with memoize must
    be hashable.

    If a memoized method is invoked directly on its class the result will not
    be cached. Instead the method will be invoked like a static method::

       class Obj(object):
           @memoize
           def add_to(self, arg):
               return self + arg

       Obj.add_to(1) # not enough arguments
       Obj.add_to(1, 2) # returns 3, result is not cached
    """

    def __init__(self, func):
        """Initialize a memoize object."""
        self.func = func
        update_wrapper(self, func)

    def __get__(self, obj, objtype=None):
        """Get the attribute from the object."""
        return partial(self, obj)

    def __call__(self, *args, **kw):
        """Callable for decorator."""
        obj = args[0]
        try:
            cache = obj.__cache  # pylint: disable=protected-access
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(kw.items()))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res


def _warn_deprecated(msg):
    """Issue a deprecation warning."""
    warnings.simplefilter('always', DeprecationWarning)
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    warnings.simplefilter('default', DeprecationWarning)


def deprecated(fun_name=None, msg=""):
    """Issue a deprecation warning for a function."""
    def _deprecated(fun):
        """Issue a deprecation warning for a function."""
        @wraps(fun)
        def _wrapper(*args, **kwargs):
            """Issue deprecation warning and forward arguments to fun."""
            name = fun_name if fun_name is not None else fun.__name__
            _warn_deprecated('Call to deprecated function %s. %s' % (name, msg))
            return fun(*args, **kwargs)

        return _wrapper

    return _deprecated


def deprecated_module(mod_name, msg=""):
    """Issue a deprecation warning for a module."""
    _warn_deprecated('Module %s is deprecated. %s' % (mod_name, msg))


class NeuromJSON(json.JSONEncoder):
    """JSON encoder that handles numpy types.

    In python3, numpy.dtypes don't serialize to correctly, so a custom
    converter is needed.
    """

    def default(self, o):  # pylint: disable=method-hidden
        """Override default method for numpy types."""
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        return json.JSONEncoder.default(self, o)


# pylint: disable=comparison-with-callable
class OrderedEnum(Enum):
    """An ordered enum class.

    Implementation taken here https://docs.python.org/3/library/enum.html#orderedenum

    Fixes https://github.com/BlueBrain/NeuroM/issues/697
    """

    def __ge__(self, other):
        """Check greater than or equal to."""
        if self.__class__ is other.__class__:
            return self.value >= other.value
        raise NotImplementedError

    def __gt__(self, other):
        """Check greater than."""
        if self.__class__ is other.__class__:
            return self.value > other.value
        raise NotImplementedError

    def __le__(self, other):
        """Check less than or equal to."""
        if self.__class__ is other.__class__:
            return self.value <= other.value
        raise NotImplementedError

    def __lt__(self, other):
        """Check less than."""
        if self.__class__ is other.__class__:
            return self.value < other.value
        raise NotImplementedError
